"""
This is a configuration file for `rich_click`, used to customize the
visual aesthetic and output of the CLI.

This is the list of all documented configuration options for `rich_click`.

You can set a style attribute by adding one or more of the following words:

- "bold" or "b" for bold text.
- "blink" for text that flashes (use this one sparingly).
- "blink2" for text that flashes rapidly (not supported by most terminals).
- "conceal" for concealed text (not supported by most terminals).
- "italic" or "i" for italic text (not supported on Windows).
- "reverse" or "r" for text with foreground and background colors reversed.
- "strike" or "s" for text with a line through it.
- "underline" or "u" for underlined text.
- "underline2" or "uu" for doubly underlined text.
- "frame" for framed text.
- "encircle" for encircled text.
- "overline" or "o" for overlined text.

The list of valid colors is here:
- https://rich.readthedocs.io/en/stable/appendix/colors.html
"""

from rich_click import rich_click

# Default styles
rich_click.STYLE_OPTION = "bold #fca311"
rich_click.STYLE_SWITCH = "bold #fca311"
rich_click.STYLE_ARGUMENT = "bold #00ff5f"
rich_click.STYLE_METAVAR = "bold #8700af"
rich_click.STYLE_METAVAR_APPEND = "dim yellow"
rich_click.STYLE_METAVAR_SEPARATOR = "dim"
rich_click.STYLE_HEADER_TEXT = ""
rich_click.STYLE_FOOTER_TEXT = ""
rich_click.STYLE_USAGE = "yellow"
rich_click.STYLE_USAGE_COMMAND = "bold"
rich_click.STYLE_DEPRECATED = "red"
rich_click.STYLE_HELPTEXT_FIRST_LINE = ""
rich_click.STYLE_HELPTEXT = ""
rich_click.STYLE_OPTION_HELP = ""
rich_click.STYLE_OPTION_DEFAULT = "bold #8700af"
rich_click.STYLE_OPTION_ENVVAR = "dim yellow"
rich_click.STYLE_REQUIRED_SHORT = "red"
rich_click.STYLE_REQUIRED_LONG = "dim red"
rich_click.STYLE_OPTIONS_PANEL_BORDER = "dim"
rich_click.ALIGN_OPTIONS_PANEL = "left"
rich_click.STYLE_OPTIONS_TABLE_SHOW_LINES = False
rich_click.STYLE_OPTIONS_TABLE_LEADING = 0
rich_click.STYLE_OPTIONS_TABLE_PAD_EDGE = False
rich_click.STYLE_OPTIONS_TABLE_PADDING = (0, 1)
rich_click.STYLE_OPTIONS_TABLE_BOX = ""
rich_click.STYLE_OPTIONS_TABLE_ROW_STYLES = None
rich_click.STYLE_OPTIONS_TABLE_BORDER_STYLE = None
rich_click.STYLE_COMMANDS_PANEL_BORDER = "dim"
rich_click.ALIGN_COMMANDS_PANEL = "left"
rich_click.STYLE_COMMANDS_TABLE_SHOW_LINES = False
rich_click.STYLE_COMMANDS_TABLE_LEADING = 0
rich_click.STYLE_COMMANDS_TABLE_PAD_EDGE = False
rich_click.STYLE_COMMANDS_TABLE_PADDING = (0, 1)
rich_click.STYLE_COMMANDS_TABLE_BOX = ""
rich_click.STYLE_COMMANDS_TABLE_ROW_STYLES = None
rich_click.STYLE_COMMANDS_TABLE_BORDER_STYLE = None
rich_click.STYLE_ERRORS_PANEL_BORDER = "red"
rich_click.ALIGN_ERRORS_PANEL = "left"
rich_click.STYLE_ERRORS_SUGGESTION = "dim"
rich_click.STYLE_ABORTED = "red"
rich_click.MAX_WIDTH = None  # Set to an int to limit to that many characters
rich_click.COLOR_SYSTEM = "auto"  # Set to None to disable colors

# Fixed strings
rich_click.HEADER_TEXT = None
rich_click.FOOTER_TEXT = None
rich_click.DEPRECATED_STRING = "(Deprecated) "
rich_click.DEFAULT_STRING = "[default: {}]"
rich_click.ENVVAR_STRING = "[env var: {}]"
rich_click.REQUIRED_SHORT_STRING = "*"
rich_click.REQUIRED_LONG_STRING = "[required]"
rich_click.RANGE_STRING = " [{}]"
rich_click.APPEND_METAVARS_HELP_STRING = "({})"
rich_click.ARGUMENTS_PANEL_TITLE = "Arguments"
rich_click.OPTIONS_PANEL_TITLE = "Options"
rich_click.COMMANDS_PANEL_TITLE = "Commands"
rich_click.ERRORS_PANEL_TITLE = "Error"
rich_click.ERRORS_SUGGESTION = (
    None  # Default: Try 'cmd -h' for help. Set to False to disable.
)
rich_click.ERRORS_EPILOGUE = None
rich_click.ABORTED_TEXT = "Aborted."

# Behaviours
rich_click.SHOW_ARGUMENTS = True  # Show positional arguments
rich_click.SHOW_METAVARS_COLUMN = (
    True  # Show a column with the option metavar (eg. INTEGER)
)
rich_click.APPEND_METAVARS_HELP = (
    False  # Append metavar (eg. [TEXT]) after the help text
)
rich_click.GROUP_ARGUMENTS_OPTIONS = (
    False  # Show arguments with options instead of in own panel
)
rich_click.USE_MARKDOWN = True  # Parse help strings as markdown
rich_click.USE_MARKDOWN_EMOJI = True  # Parse emoji codes in markdown :smile:
rich_click.USE_RICH_MARKUP = (
    False  # Parse help strings for rich markup (eg. [red]my text[/])
)
rich_click.COMMAND_GROUPS = {}  # Define sorted groups of panels to display subcommands
rich_click.OPTION_GROUPS = (
    {}
)  # Define sorted groups of panels to display options and arguments
rich_click.USE_CLICK_SHORT_HELP = (
    False  # Use click's default function to truncate help text
)



"""table updates for user invite

Revision ID: 31493e48c1d8
Revises: 4fb779cc5f17
Create Date: 2024-03-08 18:05:31.392727

"""

import sqlalchemy as sa
from alembic import op
from citext import CIText

# revision identifiers, used by Alembic.
revision = "31493e48c1d8"
down_revision = "4fb779cc5f17"
branch_labels = None
depends_on = None


def upgrade():
    op.execute("create type disabledreason as enum('pending_invite')")
    op.create_table(
        "fides_user_invite",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("username", CIText(), nullable=False),
        sa.Column("hashed_invite_code", sa.String(), nullable=False),
        sa.Column("salt", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["username"], ["fidesuser.username"], ondelete="CASCADE"
        ),
        sa.PrimaryKeyConstraint("id", "username"),
    )
    op.create_index(
        op.f("ix_fides_user_invite_id"), "fides_user_invite", ["id"], unique=False
    )
    op.add_column("fidesuser", sa.Column("email_address", CIText(), nullable=True))
    op.add_column(
        "fidesuser",
        sa.Column("disabled", sa.Boolean(), nullable=False, server_default="f"),
    )
    op.add_column(
        "fidesuser",
        sa.Column(
            "disabled_reason",
            sa.Enum("pending_invite", name="disabledreason"),
            nullable=True,
        ),
    )
    op.create_unique_constraint(
        "fidesuser_email_address", "fidesuser", ["email_address"]
    )


def downgrade():
    op.drop_constraint("fidesuser_email_address", "fidesuser", type_="unique")
    op.drop_column("fidesuser", "disabled_reason")
    op.drop_column("fidesuser", "disabled")
    op.drop_column("fidesuser", "email_address")
    op.drop_index(op.f("ix_fides_user_invite_id"), table_name="fides_user_invite")
    op.drop_table("fides_user_invite")
    op.execute("drop type disabledreason")



from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy import Column, ForeignKey, String, UniqueConstraint
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import RelationshipProperty, Session, relationship

from fides.api.db.base_class import Base, FidesBase  # type: ignore[attr-defined]
from fides.api.models.privacy_notice import PrivacyNotice


class ConsentAutomation(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "plus_consent_automation"

    connection_config_id = Column(
        String,
        ForeignKey("connectionconfig.id", ondelete="CASCADE"),
        nullable=False,
        unique=True,
    )
    consentable_items: "RelationshipProperty[List[ConsentableItem]]" = relationship(
        "ConsentableItem",
        back_populates="consent_automation",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )

    @classmethod
    def create(
        cls, db: Session, *, data: Dict[str, Any], check_name: bool = False
    ) -> "ConsentAutomation":
        consentable_items = data.pop("consentable_items", [])
        consent_automation = super().create(db=db, data=data, check_name=check_name)
        link_consentable_items_to_consent_automation(
            db, consentable_items, consent_automation
        )
        return consent_automation

    def update(self, db: Session, *, data: Dict[str, Any]) -> "ConsentAutomation":
        consentable_items = data.pop("consentable_items", [])
        super().update(db=db, data=data)
        link_consentable_items_to_consent_automation(db, consentable_items, self)
        return self

    @classmethod
    def create_or_update(cls, db: Session, *, data: Dict[str, Any]) -> "ConsentAutomation":  # type: ignore[override]
        consent_automation = ConsentAutomation.filter(
            db=db,
            conditions=(
                ConsentAutomation.connection_config_id == data["connection_config_id"]
            ),
        ).first()

        if consent_automation:
            consent_automation.update(db=db, data=data)
        else:
            consent_automation = cls.create(db=db, data=data)

        return consent_automation


def link_consentable_items_to_consent_automation(
    db: Session,
    consentable_items: List[Dict[str, Any]],
    consent_automation: ConsentAutomation,
) -> None:
    """
    Takes a hierarchical list of consentable items and maps them to database items.
    Attaches the database items to the consent automation.
    """

    existing_items = {
        (item.type, str(item.external_id)): item
        for item in consent_automation.consentable_items
    }

    def process_items(
        items_data: List[Dict[str, Any]], parent_id: Optional[str] = None
    ) -> List[ConsentableItem]:
        processed_items = []
        for item_data in items_data:
            external_id = str(item_data["external_id"])
            item_type = item_data["type"]

            key = (item_type, external_id)
            item = existing_items.get(key)

            if item:
                item.notice_id = item_data.get("notice_id")
            else:
                item = ConsentableItem(
                    external_id=external_id,
                    consent_automation_id=consent_automation.id,
                    parent_id=parent_id,
                    type=item_type,
                    name=item_data["name"],
                    notice_id=item_data.get("notice_id"),
                )
                db.add(item)
                # flush to the DB so we can get the auto-generated ID for this item
                db.flush()

            processed_items.append(item)

            if "children" in item_data:
                processed_items.extend(
                    process_items(item_data["children"], parent_id=item.id)
                )

        return processed_items

    try:
        consent_automation.consentable_items = process_items(consentable_items)
        db.commit()
    except IntegrityError as exc:
        logger.error("Error occurred while attempting to save consentable items", exc)

    db.refresh(consent_automation)


class ConsentableItem(Base):
    @declared_attr
    def __tablename__(self) -> str:
        return "plus_consentable_item"

    id = Column(
        String(255), primary_key=True, index=True, default=FidesBase.generate_uuid
    )

    consent_automation_id = Column(
        String,
        ForeignKey(ConsentAutomation.id_field_path, ondelete="CASCADE"),
        nullable=False,
    )
    external_id = Column(
        String,
        nullable=False,
    )
    parent_id = Column(
        String,
        ForeignKey("plus_consentable_item.id", ondelete="CASCADE"),
        nullable=True,
    )
    notice_id = Column(String, ForeignKey(PrivacyNotice.id_field_path), nullable=True)
    type = Column(
        String,
        nullable=False,
    )
    name = Column(String, nullable=False)
    consent_automation = relationship(
        "ConsentAutomation", back_populates="consentable_items"
    )
    children: "RelationshipProperty[List[ConsentableItem]]" = relationship(
        "ConsentableItem",
        back_populates="parent",
        cascade="all, delete-orphan",
        passive_deletes=True,
    )
    parent: "RelationshipProperty[Optional[ConsentableItem]]" = relationship(
        "ConsentableItem",
        back_populates="children",
        remote_side=[id],
    )

    __table_args__ = (UniqueConstraint("consent_automation_id", "type", "external_id"),)




# This file was generated by 'versioneer.py' (0.19) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2025-04-16T11:42:20+0200",
 "dirty": true,
 "error": null,
 "full-revisionid": "943a31433f896fe25cd9884d836cc029465c1e3b",
 "version": "2.59.1"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



from jose import jwe
from jose.constants import ALGORITHMS

_JWT_ENCRYPTION_ALGORITHM = ALGORITHMS.A256GCM


def generate_jwe(payload: str, encryption_key: str, encoding: str = "UTF-8") -> str:
    """Generates a JWE with the provided payload.

    Returns a string representation.
    """
    return jwe.encrypt(
        payload,
        encryption_key,
        encryption=_JWT_ENCRYPTION_ALGORITHM,
    ).decode(encoding)


def decrypt_jwe(token: str, encryption_key: str, encoding: str = "UTF-8") -> str:
    """Decrypts a JWE token and returns the payload as a string.

    Args:
        token: The JWE token to decrypt.
        encryption_key: The key used to decrypt the token.
        encoding: The encoding to use for the decrypted payload.

    Returns:
        The decrypted payload as a string.
    """
    decrypted_payload = jwe.decrypt(
        token,
        encryption_key,
    )
    return decrypted_payload.decode(encoding)



import json
from typing import Dict, List, Optional

from loguru import logger
from sqlalchemy import text
from sqlalchemy.engine import Connection, ResultProxy
from sqlalchemy.sql.elements import TextClause

from fides.api.alembic.migrations.helpers.database_functions import generate_record_id
from fides.api.db.seed import DEFAULT_ACCESS_POLICY_RULE, DEFAULT_ERASURE_POLICY_RULE
from fides.api.schemas.policy import ActionType
from fides.api.util.text import to_snake_case


def _replace_matching_data_label(
    data_label: str, data_label_map: Dict[str, str]
) -> str:
    """
    Helper function to do string replacement for updated fides_keys.
    """
    for old, new in data_label_map.items():
        if data_label and data_label.startswith(old):
            return data_label.replace(old, new)

    return data_label


def update_privacy_declarations(
    bind: Connection, data_use_map: Dict[str, str], data_category_map: Dict[str, str]
) -> None:
    """
    Upgrade or downgrade Privacy Declarations for Fideslang 2.0

    This updates:
    - data uses
    - data categories
    - shared categories
    """
    existing_privacy_declarations: ResultProxy = bind.execute(
        text(
            "SELECT id, data_use, data_categories, shared_categories FROM privacydeclaration;"
        )
    )
    for row in existing_privacy_declarations:
        data_use: str = _replace_matching_data_label(row["data_use"], data_use_map)
        data_categories: List[str] = [
            _replace_matching_data_label(data_category, data_category_map)
            for data_category in row["data_categories"]
        ]
        shared_categories: List[str] = [
            _replace_matching_data_label(data_category, data_category_map)
            for data_category in row["shared_categories"]
        ]

        update_query: TextClause = text(
            "UPDATE privacydeclaration SET data_use = :updated_use, data_categories = :updated_categories, shared_categories = :updated_shared WHERE id= :declaration_id"
        )
        bind.execute(
            update_query,
            {
                "declaration_id": row["id"],
                "updated_use": data_use,
                "updated_categories": data_categories,
                "updated_shared": shared_categories,
            },
        )


def update_ctl_policies(
    bind: Connection, data_use_map: Dict[str, str], data_category_map: Dict[str, str]
) -> None:
    """
    Upgrade or downgrade Policy Rules for Fideslang 2.0

    This updates:
    - data uses
    - data categories
    """
    existing_ctl_policies: ResultProxy = bind.execute(
        text("SELECT id, rules FROM ctl_policies;")
    )

    for row in existing_ctl_policies:
        rules: List[Dict] = row["rules"]

        for i, rule in enumerate(rules or []):
            data_uses: List = rule.get("data_uses", {}).get("values", [])
            rules[i]["data_uses"]["values"] = [
                _replace_matching_data_label(use, data_use_map) for use in data_uses
            ]

            data_categories: List = rule.get("data_categories", {}).get("values", [])
            rules[i]["data_categories"]["values"] = [
                _replace_matching_data_label(category, data_category_map)
                for category in data_categories
            ]

        update_data_use_query: TextClause = text(
            "UPDATE ctl_policies SET rules = :updated_rules WHERE id= :policy_id"
        )
        bind.execute(
            update_data_use_query,
            {"policy_id": row["id"], "updated_rules": json.dumps(rules)},
        )


def update_data_label_tables(
    bind: Connection, update_map: Dict[str, str], table_name: str
) -> None:
    """
    Upgrade or downgrade Data Labels for Fideslang 2.0
    """
    existing_labels: ResultProxy = bind.execute(
        text(f"SELECT fides_key, parent_key FROM {table_name};")
    )
    for row in existing_labels:
        old_key = row["fides_key"]
        new_key = _replace_matching_data_label(old_key, update_map)

        old_parent = row["parent_key"]
        new_parent = _replace_matching_data_label(old_parent, update_map)

        update_query: TextClause = text(
            f"UPDATE {table_name} SET fides_key = :updated_key, parent_key = :updated_parent WHERE fides_key = :old_key"
        )
        bind.execute(
            update_query,
            {
                "updated_key": new_key,
                "old_key": old_key,
                "updated_parent": new_parent,
            },
        )


def update_rule_targets(bind: Connection, data_label_map: Dict[str, str]) -> None:
    """Upgrade ruletargets to use the new data categories."""

    existing_rule_targets: ResultProxy = bind.execute(
        text("SELECT id, data_category FROM ruletarget;")
    )

    for row in existing_rule_targets:
        data_category = row["data_category"]

        if not data_category:
            continue

        updated_category: str = _replace_matching_data_label(
            data_category, data_label_map
        )

        update_data_category_query: TextClause = text(
            "UPDATE ruletarget SET data_category = :updated_category WHERE id= :target_id"
        )
        bind.execute(
            update_data_category_query,
            {"target_id": row["id"], "updated_category": updated_category},
        )


def update_datasets_data_categories(
    bind: Connection, data_label_map: Dict[str, str]
) -> None:
    """Upgrade the datasets and their collections/fields in the database to use the new data categories."""

    # Get all datasets out of the database
    existing_datasets: ResultProxy = bind.execute(
        text("SELECT id, data_categories, collections FROM ctl_datasets;")
    )

    for row in existing_datasets:
        # Update data categories at the top level
        dataset_data_categories: Optional[List[str]] = row["data_categories"]

        if dataset_data_categories:
            updated_categories: List[str] = [
                _replace_matching_data_label(category, data_label_map)
                for category in dataset_data_categories
            ]

            update_label_query: TextClause = text(
                "UPDATE ctl_datasets SET data_categories = :updated_labels WHERE id= :dataset_id"
            )
            bind.execute(
                update_label_query,
                {"dataset_id": row["id"], "updated_labels": updated_categories},
            )

        # Update the collections objects
        collections: str = json.dumps(row["collections"])
        if collections:
            for key, value in data_label_map.items():
                collections = collections.replace(key, value)

            update_collections_query: TextClause = text(
                "UPDATE ctl_datasets SET collections = :updated_collections WHERE id= :dataset_id"
            )
            bind.execute(
                update_collections_query,
                {"dataset_id": row["id"], "updated_collections": collections},
            )


def update_system_ingress_egress_data_categories(
    bind: Connection, data_label_map: Dict[str, str]
) -> None:
    """Upgrade or downgrade data categories on system DataFlow objects (egress/ingress)"""
    existing_systems: ResultProxy = bind.execute(
        text("SELECT id, egress, ingress FROM ctl_systems;")
    )

    for row in existing_systems:
        ingress = row["ingress"]
        egress = row["egress"]

        # Do a blunt find/replace
        if ingress:
            for item in ingress:
                if item["data_categories"]:
                    item["data_categories"] = [
                        _replace_matching_data_label(category, data_label_map)
                        for category in item["data_categories"]
                    ]

            update_ingress_query: TextClause = text(
                "UPDATE ctl_systems SET ingress = :updated_ingress WHERE id= :system_id"
            )
            bind.execute(
                update_ingress_query,
                {"system_id": row["id"], "updated_ingress": json.dumps(ingress)},
            )

        if egress:
            for item in egress:
                if item["data_categories"]:
                    item["data_categories"] = [
                        _replace_matching_data_label(category, data_label_map)
                        for category in item["data_categories"]
                    ]

            update_egress_query: TextClause = text(
                "UPDATE ctl_systems SET egress = :updated_egress WHERE id= :system_id"
            )
            bind.execute(
                update_egress_query,
                {"system_id": row["id"], "updated_egress": json.dumps(egress)},
            )


def update_privacy_notices(bind: Connection, data_use_map: Dict[str, str]) -> None:
    """
    Update the Privacy Notice Models.

    This includes the following models:
    - PrivacyNotice
    - PrivacyNoticeHistory
    - PrivacyNoticeTemplate
    """
    privacy_notice_tables = [
        "privacynotice",
        "privacynoticetemplate",
        "privacynoticehistory",
    ]
    for table in privacy_notice_tables:
        existing_notices: ResultProxy = bind.execute(
            text(f"SELECT id, data_uses FROM {table};")
        )

        for row in existing_notices:
            data_uses = row["data_uses"]

            # Do a blunt find/replace
            updated_data_uses = [
                _replace_matching_data_label(use, data_use_map) for use in data_uses
            ]

            update_query: TextClause = text(
                f"UPDATE {table} SET data_uses= :updated_uses WHERE id= :notice_id"
            )
            bind.execute(
                update_query,
                {"notice_id": row["id"], "updated_uses": updated_data_uses},
            )


def update_consent(bind: Connection, data_use_map: Dict[str, str]) -> None:
    """
    Update Consent objects in the database.
    """

    # Update the Consent table
    existing_consents: ResultProxy = bind.execute(
        text("SELECT provided_identity_id, data_use FROM consent;")
    )

    for row in existing_consents:
        updated_use: str = _replace_matching_data_label(row["data_use"], data_use_map)

        update_label_query: TextClause = text(
            "UPDATE consent SET data_use= :updated_label WHERE provided_identity_id= :key AND data_use = :old_use"
        )
        bind.execute(
            update_label_query,
            {
                "key": row["provided_identity_id"],
                "old_use": row["data_use"],
                "updated_label": updated_use,
            },
        )

    # Update the Privacy Request Table
    existing_privacy_requests: ResultProxy = bind.execute(
        text("select id, consent_preferences from privacyrequest;")
    )

    for row in existing_privacy_requests:
        preferences: List[Dict] = row["consent_preferences"]

        if preferences:
            for index, preference in enumerate(preferences):
                preferences[index]["data_use"] = _replace_matching_data_label(
                    data_label=preference["data_use"], data_label_map=data_use_map
                )

        update_pr_query: TextClause = text(
            "UPDATE privacyrequest SET consent_preferences= :updated_preferences WHERE id= :id"
        )
        bind.execute(
            update_pr_query,
            {"id": row["id"], "updated_preferences": json.dumps(preferences)},
        )

    # Update the Consent Request Table
    existing_consent_requests: ResultProxy = bind.execute(
        text("select id, preferences from consentrequest;")
    )

    for row in existing_consent_requests:
        preferences: List[Dict] = row["preferences"]

        if preferences:
            for index, preference in enumerate(preferences):
                preferences[index]["data_use"] = _replace_matching_data_label(
                    data_label=preference["data_use"], data_label_map=data_use_map
                )

        update_cr_query: TextClause = text(
            "UPDATE consentrequest SET preferences= :updated_preferences WHERE id= :id"
        )
        bind.execute(
            update_cr_query,
            {"id": row["id"], "updated_preferences": json.dumps(preferences)},
        )


def remove_conflicting_rule_targets(bind: Connection):
    """
    Iterates through all of the erasure policies and removes level 3 data categories in favor of level 2 data categories.

    For example: user.demographic is preserved over user.demographic.*

    This is needed because RuleTarget.create() validates all sibling rule targets to prevent invalid masking scenarios.
    """
    erasure_rules: ResultProxy = bind.execute(
        text("SELECT id, key FROM rule WHERE action_type = :action_type"),
        {"action_type": ActionType.erasure.value},
    )

    for rule in erasure_rules:
        all_categories_query: ResultProxy = bind.execute(
            text("SELECT data_category FROM ruletarget WHERE rule_id = :rule_id"),
            {"rule_id": rule.id},
        )
        all_categories = {row.data_category for row in all_categories_query}

        rule_targets = bind.execute(
            text("SELECT id, data_category FROM ruletarget WHERE rule_id = :rule_id"),
            {"rule_id": rule.id},
        )

        rule_targets_to_remove = []
        for target in rule_targets:
            parts = target.data_category.split(".")
            if len(parts) == 3:
                parent_category = f"{parts[0]}.{parts[1]}"
                if parent_category in all_categories:
                    rule_targets_to_remove.append(target)
                    logger.info(
                        f"Marking conflicting rule target {target.data_category} for removal from rule {rule.key}"
                    )

        if rule_targets_to_remove:
            target_ids = [target.id for target in rule_targets_to_remove]
            bind.execute(
                text("DELETE FROM ruletarget WHERE id IN :target_ids"),
                {"target_ids": tuple(target_ids)},
            )
            logger.info(f"Removed {len(target_ids)} conflicting rule targets")


def update_default_dsr_policies(bind: Connection) -> None:
    """
    Updates the default policies with new data categories using manual insertion.
    """

    new_data_categories = [
        "user.behavior",
        "user.content",
        "user.privacy_preferences",
    ]

    rules: ResultProxy = bind.execute(
        text(
            "SELECT id, key FROM rule WHERE key IN (:access_policy, :erasure_policy);"
        ),
        {
            "access_policy": DEFAULT_ACCESS_POLICY_RULE,
            "erasure_policy": DEFAULT_ERASURE_POLICY_RULE,
        },
    )

    if rules.rowcount == 0:
        logger.info("No default policies were found to update")
        return

    updates_made = False
    for rule in rules:
        for data_category in new_data_categories:
            compound_key = to_snake_case(f"{rule.id}_{data_category}")

            # check if the rule target already exists
            existing_target: ResultProxy = bind.execute(
                text(
                    "SELECT 1 FROM ruletarget WHERE rule_id = :rule_id AND data_category = :data_category"
                ),
                {"rule_id": rule.id, "data_category": data_category},
            ).first()

            if existing_target is None:
                # Insert rule targets directly into the database to bypass validation checks.
                # Invalid entries are removed in remove_conflicting_rule_targets
                bind.execute(
                    text(
                        "INSERT INTO ruletarget (id, name, key, data_category, rule_id) VALUES (:id, :name, :key, :data_category, :rule_id)"
                    ),
                    {
                        "id": generate_record_id("rul"),
                        "name": f"{rule.id}-{data_category}",
                        "key": compound_key,
                        "data_category": data_category,
                        "rule_id": rule.id,
                    },
                )
                logger.info(
                    f"Inserted new rule target: {data_category} for rule {rule.key}"
                )
                updates_made = True
            else:
                logger.info(
                    f"Rule target already exists: {data_category} for rule {rule.key}"
                )

    if updates_made:
        logger.info("The default policies have been updated with new data categories")
    else:
        logger.info("No updates were necessary for the default policies")



"""initial

Revision ID: 0210948a8147
Revises: 
Create Date: 2021-10-29 11:51:19.265141

"""

import sqlalchemy as sa
import sqlalchemy_utils
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "0210948a8147"
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "client",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("hashed_secret", sa.String(), nullable=False),
        sa.Column("salt", sa.String(), nullable=False),
        sa.Column("scopes", sa.ARRAY(sa.String()), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_client_id"), "client", ["id"], unique=False)
    op.create_table(
        "connectionconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column(
            "connection_type",
            sa.Enum("postgres", "mongodb", "mysql", "https", name="connectiontype"),
            nullable=False,
        ),
        sa.Column(
            "access", sa.Enum("read", "write", name="accesslevel"), nullable=False
        ),
        sa.Column(
            "secrets",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("last_test_timestamp", sa.DateTime(timezone=True), nullable=True),
        sa.Column("last_test_succeeded", sa.Boolean(), nullable=True),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_connectionconfig_id"), "connectionconfig", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_connectionconfig_key"), "connectionconfig", ["key"], unique=True
    )
    op.create_index(
        op.f("ix_connectionconfig_name"), "connectionconfig", ["name"], unique=True
    )
    op.create_table(
        "executionlog",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("dataset_name", sa.String(), nullable=True),
        sa.Column("collection_name", sa.String(), nullable=True),
        sa.Column(
            "fields_affected", postgresql.JSONB(astext_type=sa.Text()), nullable=True
        ),
        sa.Column("message", sa.String(), nullable=True),
        sa.Column(
            "action_type",
            sa.Enum("access", "consent", "erasure", "update", name="actiontype"),
            nullable=False,
        ),
        sa.Column(
            "status",
            sa.Enum(
                "in_processing",
                "pending",
                "complete",
                "error",
                "retrying",
                name="executionlogstatus",
            ),
            nullable=False,
        ),
        sa.Column("privacy_request_id", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_executionlog_action_type"),
        "executionlog",
        ["action_type"],
        unique=False,
    )
    op.create_index(
        op.f("ix_executionlog_collection_name"),
        "executionlog",
        ["collection_name"],
        unique=False,
    )
    op.create_index(
        op.f("ix_executionlog_dataset_name"),
        "executionlog",
        ["dataset_name"],
        unique=False,
    )
    op.create_index(op.f("ix_executionlog_id"), "executionlog", ["id"], unique=False)
    op.create_index(
        op.f("ix_executionlog_privacy_request_id"),
        "executionlog",
        ["privacy_request_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_executionlog_status"), "executionlog", ["status"], unique=False
    )
    op.create_table(
        "storageconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=True),
        sa.Column(
            "type",
            sa.Enum(
                "s3",
                "gcs",
                "transcend",
                "onetrust",
                "ethyca",
                "local",
                name="storagetype",
            ),
            nullable=False,
        ),
        sa.Column("details", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column(
            "secrets",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column(
            "format",
            sa.Enum("json", "csv", name="responseformat"),
            server_default="json",
            nullable=False,
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(op.f("ix_storageconfig_id"), "storageconfig", ["id"], unique=False)
    op.create_index(op.f("ix_storageconfig_key"), "storageconfig", ["key"], unique=True)
    op.create_index(
        op.f("ix_storageconfig_name"), "storageconfig", ["name"], unique=True
    )
    op.create_index(
        op.f("ix_storageconfig_type"), "storageconfig", ["type"], unique=False
    )
    op.create_table(
        "datasetconfig",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.Column("fides_key", sa.String(), nullable=False),
        sa.Column("dataset", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_datasetconfig_fides_key"), "datasetconfig", ["fides_key"], unique=True
    )
    op.create_index(op.f("ix_datasetconfig_id"), "datasetconfig", ["id"], unique=False)
    op.create_table(
        "policy",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("client_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(op.f("ix_policy_id"), "policy", ["id"], unique=False)
    op.create_index(op.f("ix_policy_key"), "policy", ["key"], unique=True)
    op.create_table(
        "privacyrequest",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("external_id", sa.String(), nullable=True),
        sa.Column("started_processing_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("finished_processing_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column("requested_at", sa.DateTime(timezone=True), nullable=True),
        sa.Column(
            "status",
            sa.Enum(
                "in_processing",
                "pending",
                "complete",
                "error",
                name="privacyrequeststatus",
            ),
            nullable=False,
        ),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.Column("origin", sa.String(), nullable=True),
        sa.Column("policy_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["policy_id"],
            ["policy.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_privacyrequest_external_id"),
        "privacyrequest",
        ["external_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_privacyrequest_id"), "privacyrequest", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_privacyrequest_status"), "privacyrequest", ["status"], unique=False
    )
    op.create_table(
        "rule",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("policy_id", sa.String(), nullable=False),
        sa.Column(
            "action_type",
            sa.Enum("access", "consent", "erasure", "update", name="actiontype"),
            nullable=False,
        ),
        sa.Column(
            "masking_strategy",
            sqlalchemy_utils.types.encrypted.encrypted_type.StringEncryptedType(),
            nullable=True,
        ),
        sa.Column("storage_destination_id", sa.String(), nullable=True),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["policy_id"],
            ["policy.id"],
        ),
        sa.ForeignKeyConstraint(
            ["storage_destination_id"],
            ["storageconfig.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(op.f("ix_rule_id"), "rule", ["id"], unique=False)
    op.create_index(op.f("ix_rule_key"), "rule", ["key"], unique=True)
    op.create_table(
        "ruletarget",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("data_category", sa.String(), nullable=False),
        sa.Column("rule_id", sa.String(), nullable=False),
        sa.Column("client_id", sa.String(), nullable=True),
        sa.ForeignKeyConstraint(
            ["client_id"],
            ["client.id"],
        ),
        sa.ForeignKeyConstraint(
            ["rule_id"],
            ["rule.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
        sa.UniqueConstraint(
            "rule_id", "data_category", name="_rule_id_data_category_uc"
        ),
    )
    op.create_index(op.f("ix_ruletarget_id"), "ruletarget", ["id"], unique=False)
    op.create_index(op.f("ix_ruletarget_key"), "ruletarget", ["key"], unique=True)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_ruletarget_key"), table_name="ruletarget")
    op.drop_index(op.f("ix_ruletarget_id"), table_name="ruletarget")
    op.drop_table("ruletarget")
    op.drop_index(op.f("ix_rule_key"), table_name="rule")
    op.drop_index(op.f("ix_rule_id"), table_name="rule")
    op.drop_table("rule")
    op.drop_index(op.f("ix_privacyrequest_status"), table_name="privacyrequest")
    op.drop_index(op.f("ix_privacyrequest_id"), table_name="privacyrequest")
    op.drop_index(op.f("ix_privacyrequest_external_id"), table_name="privacyrequest")
    op.drop_table("privacyrequest")
    op.drop_index(op.f("ix_policy_key"), table_name="policy")
    op.drop_index(op.f("ix_policy_id"), table_name="policy")
    op.drop_table("policy")
    op.drop_index(op.f("ix_datasetconfig_id"), table_name="datasetconfig")
    op.drop_index(op.f("ix_datasetconfig_fides_key"), table_name="datasetconfig")
    op.drop_table("datasetconfig")
    op.drop_index(op.f("ix_storageconfig_type"), table_name="storageconfig")
    op.drop_index(op.f("ix_storageconfig_name"), table_name="storageconfig")
    op.drop_index(op.f("ix_storageconfig_key"), table_name="storageconfig")
    op.drop_index(op.f("ix_storageconfig_id"), table_name="storageconfig")
    op.drop_table("storageconfig")
    op.drop_index(op.f("ix_executionlog_status"), table_name="executionlog")
    op.drop_index(op.f("ix_executionlog_privacy_request_id"), table_name="executionlog")
    op.drop_index(op.f("ix_executionlog_id"), table_name="executionlog")
    op.drop_index(op.f("ix_executionlog_dataset_name"), table_name="executionlog")
    op.drop_index(op.f("ix_executionlog_collection_name"), table_name="executionlog")
    op.drop_index(op.f("ix_executionlog_action_type"), table_name="executionlog")
    op.drop_table("executionlog")
    op.drop_index(op.f("ix_connectionconfig_name"), table_name="connectionconfig")
    op.drop_index(op.f("ix_connectionconfig_key"), table_name="connectionconfig")
    op.drop_index(op.f("ix_connectionconfig_id"), table_name="connectionconfig")
    op.drop_table("connectionconfig")
    op.drop_index(op.f("ix_client_id"), table_name="client")
    op.drop_table("client")
    # ### end Alembic commands ###



"""drop key column from property

Revision ID: 3815b0db5b14
Revises: 32497f08e227
Create Date: 2024-02-29 21:54:38.751678

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3815b0db5b14"
down_revision = "32497f08e227"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_index("ix_plus_property_key", table_name="plus_property")
    op.drop_column("plus_property", "key")


def downgrade():
    op.add_column(
        "plus_property",
        sa.Column("key", sa.VARCHAR(), autoincrement=False, nullable=True),
    )
    op.create_index("ix_plus_property_key", "plus_property", ["key"], unique=False)



"""cookie_system_unique_constraint

Revision ID: 3cc39a6ce32b
Revises: 842790fa918a
Create Date: 2023-11-07 18:07:42.861201

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
from sqlalchemy import text

revision = "3cc39a6ce32b"
down_revision = "842790fa918a"
branch_labels = None
depends_on = None


def upgrade():
    bind = op.get_bind()

    # Remove duplicate cookies on system id and system name first
    bind.execute(
        text(
            "DELETE FROM cookies A USING cookies B "
            "WHERE A.id < B.id AND A.name = B.name "
            "AND A.system_id = B.system_id "
            "AND A.system_id IS NOT NULL"
        )
    )

    # Add unique cookie on name and system id
    op.create_unique_constraint(
        "_cookie_name_system_uc", "cookies", ["name", "system_id"]
    )


def downgrade():
    op.drop_constraint("_cookie_name_system_uc", "cookies", type_="unique")



from datetime import datetime
from typing import Any, Dict, List, Optional, cast

from fideslang.models import Dataset
from fideslang.validation import FidesKey
from loguru import logger
from pydantic import BaseModel, ConfigDict, model_validator

from fides.api.common_exceptions import NoSuchConnectionTypeSecretSchemaError
from fides.api.models.connectionconfig import AccessLevel, ConnectionType
from fides.api.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.schemas.connection_configuration import connection_secrets_schemas
from fides.api.schemas.policy import ActionType
from fides.api.schemas.saas.saas_config import SaaSConfigBase
from fides.api.util.connection_type import get_connection_type_secret_schema


class CreateConnectionConfiguration(BaseModel):
    """
    Schema for creating a ConnectionConfiguration

    Note that secrets are *NOT* allowed to be supplied here.
    """

    name: Optional[str] = None
    key: Optional[FidesKey] = None
    connection_type: ConnectionType
    access: AccessLevel
    disabled: Optional[bool] = False
    description: Optional[str] = None
    model_config = ConfigDict(
        from_attributes=True, use_enum_values=True, extra="ignore"
    )


class CreateConnectionConfigurationWithSecrets(CreateConnectionConfiguration):
    """Schema for creating a connection configuration including secrets."""

    secrets: Optional[connection_secrets_schemas] = None
    saas_connector_type: Optional[str] = None
    model_config = ConfigDict(from_attributes=True, extra="ignore")


def mask_sensitive_fields(
    connection_secrets: Dict[str, Any], secret_schema: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Mask sensitive fields in the given secrets based on the provided schema.
    This function traverses the given secrets dictionary and uses the provided schema to
    identify fields that have been marked as sensitive. The function replaces the sensitive
    field values with a mask string ('********').
    Args:
        connection_secrets (Dict[str, Any]): The secrets to be masked.
        secret_schema (Dict[str, Any]): The schema defining which fields are sensitive.
    Returns:
        Dict[str, Any]: The secrets dictionary with sensitive fields masked.
    """
    if connection_secrets is None:
        return connection_secrets

    secret_schema_properties = secret_schema["properties"]
    new_connection_secrets = {}

    for key, value in connection_secrets.items():
        if key in secret_schema_properties:
            if secret_schema_properties.get(key, {}).get("sensitive", False):
                new_connection_secrets[key] = "**********"
            else:
                new_connection_secrets[key] = value

    return new_connection_secrets


class ConnectionConfigSecretsMixin(BaseModel):
    """
    A schema mixin to declare a connection config `secrets` attribute
    and handle masking of sensitive values based on `connection_type`
    and (optionally) a `saas_config`.
    """

    connection_type: ConnectionType
    secrets: Optional[Dict[str, Any]] = None
    saas_config: Optional[SaaSConfigBase] = None

    @model_validator(mode="after")
    def mask_sensitive_values(self) -> "ConnectionConfigSecretsMixin":
        """
        Mask sensitive values in the `secrets` attribute based on `connection_type`
        and (optionally) a `saas_config`.
        """
        if self.secrets is None:
            return self

        connection_type = (
            self.saas_config.type
            if self.connection_type == ConnectionType.saas and self.saas_config
            else self.connection_type.value  # type: ignore
        )
        try:
            secret_schema = get_connection_type_secret_schema(
                connection_type=connection_type
            )
        except NoSuchConnectionTypeSecretSchemaError as e:
            logger.error(e)
            # if there is no schema, we don't know what values to mask.
            # so all the secrets are removed.
            self.secrets = None
            return self

        self.secrets = mask_sensitive_fields(cast(dict, self.secrets), secret_schema)
        return self


class ConnectionConfigurationResponse(ConnectionConfigSecretsMixin):
    """
    Describes the returned schema for a ConnectionConfiguration.

    The mixin base class ensures that `secrets` sensitive values are masked.
    """

    name: Optional[str] = None
    key: FidesKey
    description: Optional[str] = None
    access: AccessLevel
    created_at: datetime
    updated_at: Optional[datetime] = None
    disabled: Optional[bool] = False
    last_test_timestamp: Optional[datetime] = None
    last_test_succeeded: Optional[bool] = None
    authorized: Optional[bool] = False
    enabled_actions: Optional[List[ActionType]] = None

    model_config = ConfigDict(from_attributes=True)


class BulkPutConnectionConfiguration(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfiguration responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class BulkPatchConnectionConfigurationWithSecrets(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of ConnectionConfigurationWithSecrets responses."""

    succeeded: List[ConnectionConfigurationResponse]
    failed: List[BulkUpdateFailed]


class SaasConnectionTemplateResponse(BaseModel):
    connection: ConnectionConfigurationResponse
    dataset: Dataset



from typing import Any, Dict, List, Optional

from loguru import logger
from sqlalchemy.orm import Session

from fides.api.common_exceptions import PrivacyRequestExit
from fides.api.graph.graph import DatasetGraph
from fides.api.models.connectionconfig import ConnectionConfig
from fides.api.models.policy import Policy
from fides.api.models.privacy_request import PrivacyRequest
from fides.api.schemas.policy import ActionType
from fides.api.task.create_request_tasks import (
    run_access_request,
    run_consent_request,
    run_erasure_request,
)
from fides.api.task.deprecated_graph_task import (
    run_access_request_deprecated,
    run_consent_request_deprecated,
    run_erasure_request_deprecated,
)
from fides.api.util.collection_util import Row
from fides.config import CONFIG


def use_dsr_3_0_scheduler(
    privacy_request: PrivacyRequest, action_type: ActionType
) -> bool:
    """Return whether we should use the DSR 3.0 scheduler.

    Override if we have a partially processed Privacy Request that was already run on
    DSR 2.0 so we can finish processing it on 2.0.

    """
    use_dsr_3_0 = CONFIG.execution.use_dsr_3_0

    prev_results: Dict[str, Optional[List[Row]]] = (
        privacy_request.get_raw_access_results()
    )
    existing_tasks_count: int = privacy_request.get_tasks_by_action(action_type).count()

    if prev_results and use_dsr_3_0 and not existing_tasks_count:
        # If we've previously tried to process this Privacy Request using DSR 2.0, continue doing so
        # for access and erasure requests
        logger.info(
            "Overriding scheduler to run privacy request {} using DSR 2.0 as it's "
            "already partially processed",
            privacy_request.id,
        )
        use_dsr_3_0 = False

    return use_dsr_3_0


def access_runner(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
    privacy_request_proceed: bool = True,  # Can be set to False in testing to run this in isolation
) -> Dict[str, List[Row]]:
    """
    Access runner that temporarily supports running Access Requests with either DSR 3.0 or DSR 2.0

    DSR 2.0 will be going away
    """
    use_dsr_3_0 = use_dsr_3_0_scheduler(privacy_request, ActionType.access)

    if use_dsr_3_0:
        run_access_request(
            privacy_request=privacy_request,
            policy=policy,
            graph=graph,
            connection_configs=connection_configs,
            identity=identity,
            session=session,
            privacy_request_proceed=privacy_request_proceed,
        )
        raise PrivacyRequestExit()

    return run_access_request_deprecated(
        privacy_request=privacy_request,
        policy=policy,
        graph=graph,
        connection_configs=connection_configs,
        identity=identity,
        session=session,
    )


def erasure_runner(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    access_request_data: Dict[str, List[Row]],
    session: Session,
    privacy_request_proceed: bool = True,  # Can be set to False in testing to run this in isolation
) -> Dict[str, int]:
    """Erasure runner that temporarily supports running Erasure DAGs with DSR 3.0 or 2.0.

    DSR 2.0 will be going away
    """
    use_dsr_3_0 = use_dsr_3_0_scheduler(privacy_request, ActionType.erasure)

    if use_dsr_3_0:
        run_erasure_request(
            privacy_request=privacy_request,
            session=session,
            privacy_request_proceed=privacy_request_proceed,
        )
        raise PrivacyRequestExit()

    return run_erasure_request_deprecated(
        privacy_request=privacy_request,
        policy=policy,
        graph=graph,
        connection_configs=connection_configs,
        identity=identity,
        access_request_data=access_request_data,
        session=session,
    )


def consent_runner(
    privacy_request: PrivacyRequest,
    policy: Policy,
    graph: DatasetGraph,
    connection_configs: List[ConnectionConfig],
    identity: Dict[str, Any],
    session: Session,
    privacy_request_proceed: bool = True,  # Can be set to False in testing to run this in isolation
) -> Dict[str, bool]:
    """Consent runner that temporarily supports running Consent DAGs with DSR 3.0 or 2.0.

    DSR 2.0 will be going away"""
    use_dsr_3_0 = use_dsr_3_0_scheduler(privacy_request, ActionType.consent)

    if use_dsr_3_0:
        run_consent_request(
            privacy_request=privacy_request,
            graph=graph,
            identity=identity,
            session=session,
            privacy_request_proceed=privacy_request_proceed,
        )
        raise PrivacyRequestExit()

    return run_consent_request_deprecated(
        privacy_request=privacy_request,
        policy=policy,
        graph=graph,
        connection_configs=connection_configs,
        identity=identity,
        session=session,
    )



from datetime import datetime
from typing import Any, Dict, List, Optional

from fides.api.schemas.base_class import FidesSchema


class TaskDetails(FidesSchema):
    task_id: str
    task_name: str
    args: List[Any]
    keyword_args: Optional[Dict[str, Any]] = None
    started_at: Optional[str] = None

    @classmethod
    def from_celery_task(cls, task: dict, state: str) -> "TaskDetails":
        timestamp = (
            datetime.fromtimestamp(task["time_start"]).isoformat()
            if task.get("time_start")
            else None
        )

        # Only include kwargs for execute request tasks
        keyword_args = None
        if any(
            task_type in task["name"]
            for task_type in ["run_access_node", "run_erasure_node", "run_consent_node"]
        ):
            keyword_args = task.get("kwargs", {})

        return cls(
            task_id=task["id"],
            task_name=task["name"],
            args=task.get("args", []),
            keyword_args=keyword_args,
            started_at=timestamp,
            state=state,
        )


class WorkerInfo(FidesSchema):
    active_task: Optional[TaskDetails] = None
    reserved_tasks: List[TaskDetails] = []
    registered_tasks: List[str] = []


class QueueInfo(FidesSchema):
    messages: int


class WorkerStats(FidesSchema):
    queues: Dict[str, QueueInfo]
    workers: Dict[str, WorkerInfo]



from typing import Any, Iterable, Optional, Pattern

from fastapi import FastAPI
from fastapi.middleware import Middleware
from fastapi.middleware.cors import CORSMiddleware

from fides.api.custom_types import URLOriginString


def update_cors_middleware(
    app: FastAPI,
    allow_origins: Iterable[URLOriginString],
    allow_origin_regex: Optional[Pattern[Any]],
) -> None:
    """
    Update the CORSMiddleware of the provided app with the provided origin parameters.

    In order to update CORSMiddleware after the app has already started, reversing the changes made here:
    https://github.com/encode/starlette/pull/2017/files
    """
    existing_middleware = find_cors_middleware(app)

    if existing_middleware:
        app.user_middleware.remove(existing_middleware)

    app.user_middleware.insert(
        0,
        Middleware(
            CORSMiddleware,
            allow_origins=allow_origins,
            allow_origin_regex=allow_origin_regex,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        ),
    )
    # In more recent starlette versions, you cannot add middleware after an application has started.
    # We have an endpoint that lets us update cors origins. I'm largely attempting to restore the
    # behavior starlette was originally providing here.
    app.middleware_stack = app.build_middleware_stack()


def find_cors_middleware(app: FastAPI) -> Optional[Middleware]:
    """Utility function to find the (first) cors middleware of the provided FastAPI app"""
    for mw in app.user_middleware:
        if mw.cls is CORSMiddleware:
            return mw
    return None



from typing import List

from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig


class MySQLQueryConfig(SQLQueryConfig):
    """
    Generates SQL valid for MySQL
    """

    def get_formatted_query_string(
        self,
        field_list: str,
        clauses: List[str],
    ) -> str:
        """Returns a query string with backtick formatting for tables that have the same names as
        MySQL reserved words."""
        return f'SELECT {field_list} FROM `{self.node.collection.name}` WHERE ({" OR ".join(clauses)})'



from functools import wraps
from typing import Any, Callable, Optional, TypeVar

from fastapi import HTTPException
from loguru import logger
from starlette.status import HTTP_400_BAD_REQUEST, HTTP_422_UNPROCESSABLE_ENTITY

T = TypeVar("T")


class ErrorHandler:
    """Utility class for handling errors consistently throughout the application.

    Usage Examples:
    -----------------------------------------------------------------------------

    1. Basic Validation:
    ```python
    from fastapi import FastAPI
    from fides.service.error_handling.error_handler import ErrorHandler

    app = FastAPI()

    @app.post("/users")
    def create_user(age: int):
        # Simple validation
        ErrorHandler.validate(age >= 18, "User must be 18 or older")
        return {"message": "User created"}

    @app.get("/items/{item_id}")
    def get_item(item_id: str):
        # Direct error raising
        if not item_id:
            ErrorHandler.raise_error("Item ID is required", status_code=400)
        return {"item_id": item_id}
    ```

    2. Exception Handling Decorator:
    ```python
    @app.post("/orders")
    @ErrorHandler.handle_exceptions("Failed to create order")
    def create_order(order_data: dict):
        if not order_data.get("items"):
            raise ValueError("Order must contain items")
        # Process order...
        return {"message": "Order created"}

    @app.get("/products/{product_id}")
    @ErrorHandler.handle_exceptions("Failed to fetch product", status_code=404)
    def get_product(product_id: str):
        product = database.get_product(product_id)
        if not product:
            raise ValueError("Product not found")
        return product
    ```

    3. Complex Validation:
    ```python
    @app.post("/payments")
    @ErrorHandler.handle_exceptions("Payment processing failed")
    def process_payment(payment: dict):
        # Validate amount
        ErrorHandler.validate(
            payment.get("amount", 0) > 0,
            "Payment amount must be positive",
            HTTP_400_BAD_REQUEST
        )

        # Validate currency
        ErrorHandler.validate(
            payment.get("currency") in ["USD", "EUR"],
            "Invalid currency",
            HTTP_400_BAD_REQUEST,
            "Unsupported currency provided"  # Optional log message
        )

        # Custom error for insufficient funds
        if payment.get("amount", 0) > get_balance():
            ErrorHandler.raise_error(
                "Insufficient funds",
                status_code=HTTP_400_BAD_REQUEST,
                log_message="User attempted payment exceeding balance"
            )

        return {"status": "payment processed"}
    ```

    4. Error Logging:
    ```python
    @app.post("/imports")
    @ErrorHandler.handle_exceptions("Import failed")
    def import_data(data: dict):
        try:
            process_import(data)
        except Exception as e:
            # Log error with custom message before raising
            ErrorHandler.raise_error(
                "Import failed: invalid format",
                status_code=400,
                log_message=f"Import failed with error: {str(e)}"
            )
        return {"status": "import complete"}
    ```

    Key Features:
    - Simple validation with custom error messages
    - Consistent error handling across endpoints
    - Built-in error logging
    - HTTP status code customization
    - Exception handling decorator for common patterns
    """

    @staticmethod
    def raise_error(
        detail: str,
        status_code: int = HTTP_422_UNPROCESSABLE_ENTITY,
        log_message: Optional[str] = None,
    ) -> None:
        """Raise an HTTPException with consistent logging.

        Args:
            detail: Error message to include in the HTTPException
            status_code: HTTP status code to use (default: 422)
            log_message: Optional message to log before raising the exception

        Raises:
            HTTPException: Always raised with the provided details
        """
        if log_message:
            logger.error(log_message)
        raise HTTPException(status_code=status_code, detail=detail)

    @staticmethod
    def validate(
        condition: bool,
        detail: str,
        status_code: int = HTTP_400_BAD_REQUEST,
        log_message: Optional[str] = None,
    ) -> None:
        """Validate a condition and raise an error if it's False.

        Args:
            condition: The condition to check
            detail: Error message if condition is False
            status_code: HTTP status code to use if condition is False
            log_message: Optional message to log before raising the exception

        Raises:
            HTTPException: If the condition is False
        """
        if not condition:
            ErrorHandler.raise_error(detail, status_code, log_message)

    @classmethod
    def handle_exceptions(
        cls, error_message: str, status_code: int = HTTP_422_UNPROCESSABLE_ENTITY
    ) -> Callable[[Callable[..., T]], Callable[..., T]]:
        """Decorator to handle exceptions consistently.

        Args:
            error_message: Base error message to use
            status_code: HTTP status code to use for unexpected errors

        Returns:
            Callable: The decorated function that handles exceptions

        Note:
            This decorator will catch specific exceptions and convert them to HTTPExceptions.
            HTTPExceptions are re-raised as is, while other exceptions are wrapped with
            additional context.
        """

        def decorator(func: Callable[..., T]) -> Callable[..., T]:
            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Optional[T]:
                try:
                    return func(*args, **kwargs)
                except HTTPException:
                    # Re-raise HTTP exceptions without modification
                    raise
                except (ValueError, TypeError, AttributeError, KeyError) as e:
                    # Handle common validation and data access errors
                    cls.raise_error(
                        f"{error_message}: {str(e)}",
                        status_code,
                        f"{error_message}: {e}",
                    )
                except Exception as e:  # pylint: disable=broad-except
                    # Log unexpected errors with full context but present a sanitized message
                    logger.error(
                        f"Unexpected error in {func.__name__}: {str(e)}", exc_info=True
                    )
                    cls.raise_error(
                        f"{error_message}: An unexpected error occurred",
                        status_code,
                        f"{error_message}: Unexpected {type(e).__name__}",
                    )
                return None  # This line is never reached but satisfies the return type checker

            return wrapper  # type: ignore[return-value]

        return decorator



"""Add TCF Publisher Restrictions

Revision ID: 6e565c16dae1
Revises: 67d01c4e124e
Create Date: 2025-04-02 12:35:34.105607

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "6e565c16dae1"
down_revision = "67d01c4e124e"
branch_labels = None
depends_on = None


def upgrade():
    # Create tcf_configuration table
    op.create_table(
        "tcf_configuration",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    # Create tcf_publisher_restriction table
    op.create_table(
        "tcf_publisher_restriction",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("tcf_configuration_id", sa.String(255), nullable=False),
        sa.Column("purpose_id", sa.Integer(), nullable=False),
        sa.Column(
            "restriction_type",
            sa.Enum(
                "purpose_restriction",
                "require_consent",
                "require_legitimate_interest",
                name="tcfrestrictiontype",
            ),
            nullable=False,
        ),
        sa.Column(
            "vendor_restriction",
            sa.Enum(
                "restrict_all_vendors",
                "allow_specific_vendors",
                "restrict_specific_vendors",
                name="tcfvendorrestriction",
            ),
            nullable=False,
        ),
        sa.Column(
            "range_entries",
            postgresql.ARRAY(postgresql.JSONB(astext_type=sa.Text())),
            server_default="{}",
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ["tcf_configuration_id"],
            ["tcf_configuration.id"],
            ondelete="CASCADE",
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_tcf_publisher_restriction_config_purpose"),
        "tcf_publisher_restriction",
        ["tcf_configuration_id", "purpose_id"],
        unique=False,
    )


def downgrade():
    # Drop tables
    op.drop_table("tcf_publisher_restriction")
    op.drop_table("tcf_configuration")

    # Drop enums
    op.execute("DROP TYPE tcfrestrictiontype")
    op.execute("DROP TYPE tcfvendorrestriction")



from __future__ import annotations

from typing import IO, Any, Dict, Optional, Tuple

from boto3.s3.transfer import TransferConfig
from botocore.exceptions import ClientError, ParamValidationError
from fideslang.validation import AnyHttpUrlString
from loguru import logger

from fides.api.schemas.storage.storage import StorageSecrets
from fides.api.util.aws_util import get_s3_client
from fides.config import CONFIG

LARGE_FILE_THRESHOLD = 5 * 1024 * 1024  # 5 MB threshold for large files
LARGE_FILE_WARNING_TEXT = (
    "File is too large to display directly. Please use the provided link to download."
)


def maybe_get_s3_client(
    auth_method: str, storage_secrets: Dict[StorageSecrets, Any]
) -> Any:
    """
    Returns an S3 client if the client can be created successfully, otherwise raises an exception.
    """
    try:
        return get_s3_client(auth_method, storage_secrets)
    except ClientError as e:
        logger.error(
            "Encountered error while uploading and generating link for S3 object: {}", e
        )
        raise e
    except ParamValidationError as e:
        raise ValueError(f"The parameters you provided are incorrect: {e}")


def create_presigned_url_for_s3(
    s3_client: Any, bucket_name: str, file_key: str
) -> AnyHttpUrlString:
    """
    Generates a presigned URL to share an S3 object

    :param s3_client: s3 base client
    :param bucket_name: string
    :param file_key: string
    :return: Presigned URL as string.
    """
    response = s3_client.generate_presigned_url(
        "get_object",
        Params={"Bucket": bucket_name, "Key": file_key},
        ExpiresIn=CONFIG.security.subject_request_download_link_ttl_seconds,
    )

    # The response contains the presigned URL
    return response


def generic_upload_to_s3(  # pylint: disable=R0913
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
    document: IO[bytes],
    size_threshold: int = LARGE_FILE_THRESHOLD,  # 5 MB threshold
) -> Optional[AnyHttpUrlString]:
    """
    Uploads file like objects to S3.
    Handles both small and large uploads.

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    :param document: File contents to upload
    """
    logger.info("Starting S3 Upload of {}", file_key)

    # Validate that the document is a file-like object
    if not hasattr(document, "read") or not hasattr(document, "seek"):
        raise TypeError(
            f"The 'document' parameter must be a file-like object supporting 'read' and 'seek'. "
            f"Received: {type(document)}"
        )

    # Ensure the file pointer is at the beginning
    try:
        document.seek(0)
    except Exception as e:
        raise ValueError(f"Failed to reset file pointer for document: {e}")

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)

    # Define a transfer configuration for multipart uploads
    transfer_config = TransferConfig(
        multipart_threshold=size_threshold,
        multipart_chunksize=size_threshold,
    )

    # Use upload_fileobj for efficient uploads (handles both small and large files)
    try:
        # Use upload_fileobj for efficient uploads (handles both small and large files)
        s3_client.upload_fileobj(
            Fileobj=document,
            Bucket=bucket_name,
            Key=file_key,
            Config=transfer_config,
        )
    except Exception as e:
        logger.error(f"Failed to upload file {file_key} to bucket {bucket_name}: {e}")
        raise e  # Re-raise the exception if you want it to propagate

    # Generate a presigned URL for the uploaded file
    logger.info(f"Successfully uploaded file {file_key} to bucket {bucket_name}")
    return create_presigned_url_for_s3(s3_client, bucket_name, file_key)


def generic_download_from_s3(
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
) -> Optional[AnyHttpUrlString]:
    """
    Returns a presigned URL for downloading an S3 object.

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    :return: A presigned URL (str) for the S3 object
    """
    logger.info("Generating presigned URL for downloading file: {}", file_key)

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)

    # Generate a presigned URL for the file
    return create_presigned_url_for_s3(s3_client, bucket_name, file_key)


def generic_retrieve_from_s3(
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
    size_threshold: int = LARGE_FILE_THRESHOLD,  # 5 MB threshold
) -> Tuple[str, AnyHttpUrlString]:
    """
    Retrieves arbitrary data from S3.
    Returns the file contents and presigned url to download the file if the file is small,
    or a warning message with the presigned URL to download the file if it is large.

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    :param size_threshold: Size threshold in bytes to determine small vs large files
    :return: File contents (BytesIO) for small files or a presigned URL (str) for large files
    """
    logger.info("Starting S3 Retrieve of {}", file_key)

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)
    try:
        # Get the object metadata to check its size
        head_response = s3_client.head_object(Bucket=bucket_name, Key=file_key)
        file_size = head_response["ContentLength"]

        if file_size <= size_threshold:
            # File is small, return its contents and presigned url
            response = s3_client.get_object(Bucket=bucket_name, Key=file_key)
            return response["Body"].read(), create_presigned_url_for_s3(
                s3_client, bucket_name, file_key
            )

        # File is large, return warning text and a presigned URL
        return (
            LARGE_FILE_WARNING_TEXT,
            create_presigned_url_for_s3(s3_client, bucket_name, file_key),
        )

    except Exception as e:
        logger.error("Encountered error while retrieving S3 object: {}", e)
        raise e


def generic_delete_from_s3(
    storage_secrets: Dict[StorageSecrets, Any],
    bucket_name: str,
    file_key: str,
    auth_method: str,
) -> None:
    """
    Deletes arbitrary data from s3

    :param storage_secrets: S3 storage secrets
    :param bucket_name: Name of the S3 bucket
    :param file_key: Key of the file in the bucket
    :param auth_method: Authentication method for S3
    """
    logger.info("Starting S3 Delete of {}", file_key)

    s3_client = maybe_get_s3_client(auth_method, storage_secrets)
    try:
        s3_client.delete_object(Bucket=bucket_name, Key=file_key)
    except Exception as e:
        logger.error("Encountered error while deleting s3 object: {}", e)
        raise e



"""add replaceable field to custom_connector_template

Revision ID: ff782b0dc07e
Revises: 6d6b0b7cbb36
Create Date: 2023-03-29 23:41:26.164600

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ff782b0dc07e"
down_revision = "6d6b0b7cbb36"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "custom_connector_template",
        sa.Column(
            "replaceable", sa.Boolean(), unique=False, nullable=False, default=False
        ),
    )


def downgrade():
    op.drop_column("custom_connector_template", "replaceable")



"""staged resource web monitor updates

Revision ID: 58f8edd66b69
Revises: ed96417b07d8
Create Date: 2024-12-17 19:25:36.184841

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "58f8edd66b69"
down_revision = "ed96417b07d8"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("stagedresource", sa.Column("system_id", sa.String(), nullable=True))
    op.add_column("stagedresource", sa.Column("vendor_id", sa.String(), nullable=True))
    op.create_index(
        op.f("ix_stagedresource_monitor_config_id"),
        "stagedresource",
        ["monitor_config_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_stagedresource_system_id"),
        "stagedresource",
        ["system_id"],
        unique=False,
    )
    op.create_index(
        op.f("ix_stagedresource_vendor_id"),
        "stagedresource",
        ["vendor_id"],
        unique=False,
    )
    op.create_foreign_key(None, "stagedresource", "ctl_systems", ["system_id"], ["id"])
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "stagedresource", type_="foreignkey")
    op.drop_index(op.f("ix_stagedresource_vendor_id"), table_name="stagedresource")
    op.drop_index(op.f("ix_stagedresource_system_id"), table_name="stagedresource")
    op.drop_index(
        op.f("ix_stagedresource_monitor_config_id"), table_name="stagedresource"
    )
    op.drop_column("stagedresource", "vendor_id")
    op.drop_column("stagedresource", "system_id")
    # ### end Alembic commands ###



"""
This module generates all of the routers for the boilerplate/generic
objects that don't require any extra logic.
"""

from fideslang.models import Dataset, Evaluation, Organization, Policy

from fides.api.schemas.taxonomy_extensions import DataCategory, DataSubject, DataUse

from .router_factory import generic_router_factory  # type: ignore[attr-defined]

DATA_CATEGORY_ROUTER = generic_router_factory(
    fides_model=DataCategory, model_type="data_category"
)
DATA_USE_ROUTER = generic_router_factory(fides_model=DataUse, model_type="data_use")
DATA_SUBJECT_ROUTER = generic_router_factory(
    fides_model=DataSubject, model_type="data_subject"
)
DATASET_ROUTER = generic_router_factory(fides_model=Dataset, model_type="dataset")
ORGANIZATION_ROUTER = generic_router_factory(
    fides_model=Organization, model_type="organization"
)
POLICY_ROUTER = generic_router_factory(fides_model=Policy, model_type="policy")
EVALUATION_ROUTER = generic_router_factory(
    fides_model=Evaluation, model_type="evaluation"
)



"""add errors to monitor execution

Revision ID: 3c58001ad310
Revises: bd875a8b5d96
Create Date: 2025-02-27 20:45:38.462604

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3c58001ad310"
down_revision = "bd875a8b5d96"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "monitorexecution",
        sa.Column(
            "messages",
            sa.ARRAY(sa.String()),
            nullable=False,
            server_default="{}",
        ),
    )


def downgrade() -> None:
    op.drop_column("monitorexecution", "messages")



from fides.api.schemas.base_class import FidesSchema, NoValidationSchema


class ManualWebhookSchema(FidesSchema):
    """Secrets for manual webhooks. No secrets needed at this time."""


class ManualWebhookDocsSchema(ManualWebhookSchema, NoValidationSchema):
    """Manual Webhooks Secrets Schema for API Docs"""



"""move system_dependencies to egress and remove

Revision ID: 76c02f99eec1
Revises: e798f37f0c26
Create Date: 2023-05-10 17:59:14.536666

"""

import json

import sqlalchemy as sa
from alembic import op
from sqlalchemy import text
from sqlalchemy.dialects import postgresql
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql.elements import TextClause

# revision identifiers, used by Alembic.
revision = "76c02f99eec1"
down_revision = "e798f37f0c26"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    bind = op.get_bind()
    existing_system_dependencies = bind.execute(
        text(
            "SELECT id, egress, system_dependencies FROM ctl_systems where system_dependencies is not null;"
        )
    )
    for row in existing_system_dependencies:
        row_id = row["id"]
        existing_egress = row["egress"]
        system_dependencies = row["system_dependencies"]
        updated_egress = []
        updated_egress_keys = []
        if existing_egress:
            for egress in existing_egress:
                updated_egress.append(egress)
                updated_egress_keys.append(egress["fides_key"])
        for system in system_dependencies:
            if system not in updated_egress_keys:
                egress = {
                    "fides_key": system,
                    "type": "system",
                    "data_categories": None,
                }
                updated_egress.append(egress)

        update_query: TextClause = text(
            "UPDATE ctl_systems "
            "SET egress = :updated_egress ::jsonb "
            "WHERE id = :row_id"
        )
        update_params = {}
        update_params["row_id"] = row_id
        update_params["updated_egress"] = f"{json.dumps(updated_egress)}"

        try:
            bind.execute(
                update_query,
                update_params,
            )
        except IntegrityError as exc:
            raise Exception(
                f"Fides attempted to copy ctl_systems.system_dependencies into ctl_systems.egress but got error: {exc}. "
            )
    op.drop_column("ctl_systems", "system_dependencies")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "ctl_systems",
        sa.Column(
            "system_dependencies",
            postgresql.ARRAY(sa.VARCHAR()),
            autoincrement=False,
            nullable=True,
        ),
    )
    # ### end Alembic commands ###



from typing import Any, List, Optional

from fides.api.graph.config import FieldPath
from fides.api.util.collection_util import Row


def consolidate_query_matches(
    row: Row,
    target_path: FieldPath,
    flattened_matches: Optional[List] = None,
) -> List[Any]:
    """
    Recursively consolidates values along from the target_path into a single array.

    A target_path can point to a single scalar value, an array, or even multiple values within arrays of embedded
    documents. We consolidate all values into a single flattened list, that are used in subsequent queries
    to locate records in another collection.

    :param row: Retrieved record from dataset
    :param target_path: FieldPath to applicable field
    :param flattened_matches: Recursive list where matched value(s) from the target_path are recursively added
    :return: A consolidated flattened list of matched values.
    """
    if flattened_matches is None:
        flattened_matches = []

    if isinstance(row, list):
        for elem in row:
            consolidate_query_matches(elem, target_path, flattened_matches)

    elif isinstance(row, dict):
        for key, value in row.items():
            if target_path.levels and key == target_path.levels[0]:
                consolidate_query_matches(
                    value, FieldPath(*target_path.levels[1:]), flattened_matches
                )

    elif row:
        flattened_matches.append(row)

    return flattened_matches



"""add policy webhooks

Revision ID: c5336b841d70
Revises: f206d4e7574d
Create Date: 2021-11-24 23:20:48.807643

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c5336b841d70"
down_revision = "f206d4e7574d"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "policypostwebhook",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column(
            "direction",
            sa.Enum("one_way", "two_way", name="webhookdirection"),
            nullable=False,
        ),
        sa.Column("order", sa.Integer(), nullable=False),
        sa.Column("policy_id", sa.String(), nullable=False),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.ForeignKeyConstraint(
            ["policy_id"],
            ["policy.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(
        op.f("ix_policypostwebhook_id"), "policypostwebhook", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_policypostwebhook_key"), "policypostwebhook", ["key"], unique=True
    )
    op.create_table(
        "policyprewebhook",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column(
            "direction",
            sa.Enum("one_way", "two_way", name="webhookdirection"),
            nullable=False,
        ),
        sa.Column("order", sa.Integer(), nullable=False),
        sa.Column("policy_id", sa.String(), nullable=False),
        sa.Column("connection_config_id", sa.String(), nullable=False),
        sa.ForeignKeyConstraint(
            ["connection_config_id"],
            ["connectionconfig.id"],
        ),
        sa.ForeignKeyConstraint(
            ["policy_id"],
            ["policy.id"],
        ),
        sa.PrimaryKeyConstraint("id"),
        sa.UniqueConstraint("name"),
    )
    op.create_index(
        op.f("ix_policyprewebhook_id"), "policyprewebhook", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_policyprewebhook_key"), "policyprewebhook", ["key"], unique=True
    )


def downgrade():
    op.drop_index(op.f("ix_policyprewebhook_key"), table_name="policyprewebhook")
    op.drop_index(op.f("ix_policyprewebhook_id"), table_name="policyprewebhook")
    op.drop_table("policyprewebhook")
    op.drop_index(op.f("ix_policypostwebhook_key"), table_name="policypostwebhook")
    op.drop_index(op.f("ix_policypostwebhook_id"), table_name="policypostwebhook")
    op.drop_table("policypostwebhook")



from typing import Any, Dict, Union

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from snowflake.sqlalchemy import URL as Snowflake_URL
from sqlalchemy.orm import Session

from fides.api.graph.execution import ExecutionNode
from fides.api.schemas.connection_configuration import SnowflakeSchema
from fides.api.service.connectors.query_configs.query_config import SQLQueryConfig
from fides.api.service.connectors.query_configs.snowflake_query_config import (
    SnowflakeQueryConfig,
)
from fides.api.service.connectors.sql_connector import SQLConnector
from fides.config import get_config

CONFIG = get_config()


class SnowflakeConnector(SQLConnector):
    """Connector specific to Snowflake"""

    secrets_schema = SnowflakeSchema

    def build_uri(self) -> str:
        """Build URI of format 'snowflake://<user_login_name>:<password>@<account_identifier>/<database_name>/
        <schema_name>?warehouse=<warehouse_name>&role=<role_name>'
        """
        config = self.secrets_schema(**self.configuration.secrets or {})

        kwargs = {}

        if config.account_identifier:
            kwargs["account"] = config.account_identifier
        if config.user_login_name:
            kwargs["user"] = config.user_login_name
        if config.password:
            kwargs["password"] = config.password
        if config.database_name:
            kwargs["database"] = config.database_name
        if config.schema_name:
            kwargs["schema"] = config.schema_name
        if config.warehouse_name:
            kwargs["warehouse"] = config.warehouse_name
        if config.role_name:
            kwargs["role"] = config.role_name

        url: str = Snowflake_URL(**kwargs)
        return url

    def get_connect_args(self) -> Dict[str, Any]:
        """Get connection arguments for the engine"""
        config = self.secrets_schema(**self.configuration.secrets or {})
        connect_args: Dict[str, Union[str, bytes]] = {}
        if config.private_key:
            config.private_key = config.private_key.replace("\\n", "\n")
            connect_args["private_key"] = config.private_key
            if config.private_key_passphrase:
                private_key_encoded = serialization.load_pem_private_key(
                    config.private_key.encode(),
                    password=config.private_key_passphrase.encode(),  # pylint: disable=no-member
                    backend=default_backend(),
                )
                private_key = private_key_encoded.private_bytes(
                    encoding=serialization.Encoding.DER,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=serialization.NoEncryption(),
                )
                connect_args["private_key"] = private_key
        return connect_args

    def query_config(self, node: ExecutionNode) -> SQLQueryConfig:
        """Query wrapper corresponding to the input execution_node."""

        db: Session = Session.object_session(self.configuration)
        return SnowflakeQueryConfig(
            node, SQLConnector.get_namespace_meta(db, node.address.dataset)
        )



from abc import ABC, abstractmethod
from typing import ClassVar, Dict, Optional, Set, Tuple, Type

from pydantic import BaseModel


class NamespaceMeta(BaseModel, ABC):
    """Base class for namespace metadata schemas"""

    # Registry of concrete namespace meta implementations
    _implementations: ClassVar[Dict[str, Type["NamespaceMeta"]]] = {}

    connection_type: Optional[str] = None

    def __init_subclass__(cls) -> None:
        """Register subclasses automatically when they're defined"""
        super().__init_subclass__()
        if hasattr(cls, "connection_type") and cls.connection_type:
            NamespaceMeta._implementations[cls.connection_type] = cls

    @classmethod
    def get_implementation(
        cls, connection_type: str
    ) -> Optional[Type["NamespaceMeta"]]:
        """Get the namespace meta implementation for a connection type"""
        return cls._implementations.get(connection_type)

    @classmethod
    @abstractmethod
    def get_fallback_secret_fields(cls) -> Set[Tuple]:
        """
        The required connection config secrets when namespace metadata is missing.
        Each implementation specifies which fields are required from connection secrets
        when falling back due to missing namespace metadata.
        """



"""A wrapper to make calling the API consistent across fides."""

from typing import Dict, List, Optional, Union

import requests

# Not using the constant value from fides.api.util.endpoint_utils to reduce the startup time for the CLI
API_PREFIX = "/api/v1"


def generate_resource_url(
    url: str,
    resource_type: str = "",
    resource_id: str = "",
    query_params: Optional[Union[Dict[str, str], Dict[str, List[str]]]] = None,
) -> str:
    """
    Generate a resource's URL using a base url, the resource type,
    and [optionally] the resource's ID.
    """
    base_url = f"{url}{API_PREFIX}/{resource_type}/{resource_id}"

    if not query_params:
        return base_url

    processed_query_params = []
    for key, value in query_params.items():
        if isinstance(value, list):
            processed_query_params.extend([f"{key}={v}" for v in value])
        else:
            processed_query_params.append(f"{key}={value}")

    query_string = "&".join(processed_query_params)
    return f"{url}{API_PREFIX}/{resource_type}/{resource_id}?{query_string}"


def get(
    url: str, resource_type: str, resource_id: str, headers: Dict[str, str]
) -> requests.Response:
    """
    Get a resource by its id.
    """
    resource_url = generate_resource_url(url, resource_type, resource_id)
    return requests.get(resource_url, headers=headers)


def create(
    url: str, resource_type: str, json_resource: str, headers: Dict[str, str]
) -> requests.Response:
    """
    Create a new resource.
    """
    resource_url = generate_resource_url(url, resource_type)
    return requests.post(resource_url, headers=headers, data=json_resource)


def ping(url: str) -> requests.Response:
    """
    Pings the Server on the base url to make sure it's available.
    """
    return requests.get(url)


def delete(
    url: str, resource_type: str, resource_id: str, headers: Dict[str, str]
) -> requests.Response:
    """
    Delete a resource by its id.
    """
    resource_url = generate_resource_url(url, resource_type, resource_id)
    return requests.delete(resource_url, headers=headers)


def ls(  # pylint: disable=invalid-name
    url: str,
    resource_type: str,
    headers: Dict[str, str],
    query_params: Optional[Dict[str, str]] = None,
) -> requests.Response:
    """
    Get a list of all of the resources of a certain type.
    """
    resource_url = generate_resource_url(url, resource_type, query_params=query_params)
    return requests.get(resource_url, headers=headers)


def update(
    url: str,
    resource_type: str,
    json_resource: Dict,
    headers: Dict[str, str],
) -> requests.Response:
    """
    Update an existing resource.
    """
    resource_url = generate_resource_url(url, resource_type)
    return requests.put(resource_url, headers=headers, data=json_resource)


def upsert(
    url: str,
    resource_type: str,
    resources: List[Dict],
    headers: Dict[str, str],
) -> requests.Response:
    """
    For each resource, insert the resource if it doesn't exist, update the resource otherwise.
    """

    resource_url = generate_resource_url(url, resource_type) + "upsert"
    return requests.post(resource_url, headers=headers, json=resources)


def dry_evaluate(
    url: str, resource_type: str, json_resource: str, headers: Dict[str, str]
) -> requests.Response:
    """
    Dry Evaluate a registry based on organizational policies.
    """
    resource_url = generate_resource_url(url, resource_type)
    url = f"{resource_url}evaluate/dry-run"
    return requests.post(url, headers=headers, data=json_resource)


def evaluate(
    url: str,
    resource_type: str,
    fides_key: str,
    tag: str,
    message: str,
    headers: Dict[str, str],
) -> requests.Response:
    """
    Evaluate a registry based on organizational policies.
    """
    resource_url = generate_resource_url(url, resource_type)
    url = f"{resource_url}evaluate/{fides_key}"
    return requests.get(url, headers=headers, params={"tag": tag, "message": message})


def db_action(
    server_url: str,
    headers: Dict[str, str],
    action: str,
) -> requests.Response:
    """
    Tell the API to perform a database action.
    """
    return requests.post(
        f"{server_url}{API_PREFIX}/admin/db/{action}",
        headers=headers,
        allow_redirects=False,
        timeout=30,
    )



from enum import Enum as EnumType
from typing import Any, Dict, List, Optional

from fideslang.validation import FidesKey
from pydantic import ConfigDict

from fides.api.schemas.api import BulkResponse, BulkUpdateFailed
from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.storage.storage import StorageDestinationResponse


class ActionType(str, EnumType):
    """The purpose of a particular privacy request"""

    access = "access"
    consent = "consent"
    erasure = "erasure"
    update = "update"


class CurrentStep(EnumType):
    pre_webhooks = "pre_webhooks"
    access = "access"
    upload_access = "upload_access"
    erasure = "erasure"
    finalize_erasure = "finalize_erasure"
    consent = "consent"
    finalize_consent = "finalize_consent"
    email_post_send = "email_post_send"
    post_webhooks = "post_webhooks"


# action types we actively support in policies/requests
SUPPORTED_ACTION_TYPES = {ActionType.access, ActionType.consent, ActionType.erasure}


class DrpAction(EnumType):
    """
    Enum to hold valid DRP actions. For more details, see:
    https://github.com/consumer-reports-digital-lab/data-rights-protocol#301-supported-rights-actions
    """

    access = "access"
    deletion = "deletion"
    # below are not supported
    sale_opt_out = "sale:opt_out"
    sale_opt_in = "sale:opt_in"
    access_categories = "access:categories"
    access_specific = "access:specific"


class PolicyMaskingSpec(FidesSchema):
    """Models the masking strategy definition"""

    strategy: str
    configuration: Dict[str, Any]


class PolicyMaskingSpecResponse(FidesSchema):
    """
    The schema to use when returning a masking strategy via the API. This schema omits other
    potentially sensitive fields in the masking configuration, for example the encryption
    algorithm.
    """

    strategy: str


class RuleTarget(FidesSchema):
    """An external representation of a Rule's target DataCategory within a Fidesops Policy"""

    name: Optional[str] = None
    key: Optional[FidesKey] = None
    # `data_category` is type str so that we can validate its contents against the DB records
    # outside of the schemas
    data_category: str
    model_config = ConfigDict(use_enum_values=True)


class RuleBase(FidesSchema):
    """An external representation of a Rule within a Fidesops Policy"""

    name: str
    key: Optional[FidesKey] = None
    action_type: ActionType
    model_config = ConfigDict(use_enum_values=True)


class RuleCreate(RuleBase):
    """
    The schema to use when creating a Rule. This schema accepts a storage_destination_key
    over a composite object.
    """

    storage_destination_key: Optional[FidesKey] = None
    masking_strategy: Optional[PolicyMaskingSpec] = None


class RuleResponse(RuleBase):
    """
    The schema to use when returning a Rule via the API. This schema uses a censored version
    of the `PolicyMaskingSpec` that omits the configuration to avoid exposing secrets.
    """

    storage_destination: Optional[StorageDestinationResponse] = None
    masking_strategy: Optional[PolicyMaskingSpecResponse] = None


class RuleResponseWithTargets(RuleBase):
    """
    The schema to use when returning a Rule via the API and when including the Rule's targets
    is desired. This schema uses a censored version of the `PolicyMaskingSpec` that omits the
    configuration to avoid exposing secrets.
    """

    targets: Optional[List[RuleTarget]] = None
    storage_destination: Optional[StorageDestinationResponse] = None
    masking_strategy: Optional[PolicyMaskingSpecResponse] = None


class Rule(RuleBase):
    """A representation of a Rule that features all storage destination data."""

    storage_destination: Optional[StorageDestinationResponse] = None
    masking_strategy: Optional[PolicyMaskingSpec] = None


class Policy(FidesSchema):
    """An external representation of a Fidesops Policy"""

    name: str
    key: Optional[FidesKey] = None
    drp_action: Optional[DrpAction] = None
    execution_timeframe: Optional[int] = None
    model_config = ConfigDict(use_enum_values=True, from_attributes=True)


class PolicyResponse(Policy):
    """A holistic view of a Policy record, including all foreign keys by default."""

    rules: Optional[List[RuleResponse]] = None
    drp_action: Optional[DrpAction] = None


class BulkPutRuleTargetResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of RuleTarget responses."""

    succeeded: List[RuleTarget]
    failed: List[BulkUpdateFailed]


class BulkPutRuleResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of Rule responses."""

    succeeded: List[RuleResponse]
    failed: List[BulkUpdateFailed]


class BulkPutPolicyResponse(BulkResponse):
    """Schema with mixed success/failure responses for Bulk Create/Update of Policy responses."""

    succeeded: List[PolicyResponse]
    failed: List[BulkUpdateFailed]



import json
import random
import time
from datetime import datetime
from typing import List, Optional

import jose.exceptions
from fastapi import Depends, HTTPException, Security
from fastapi_pagination import Page, Params
from fastapi_pagination.bases import AbstractPage
from fastapi_pagination.ext.sqlalchemy import paginate
from fideslang.models import System as SystemSchema
from fideslang.validation import FidesKey
from loguru import logger
from sqlalchemy.orm import Query, Session
from sqlalchemy_utils import escape_like
from starlette.status import (
    HTTP_200_OK,
    HTTP_201_CREATED,
    HTTP_204_NO_CONTENT,
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_403_FORBIDDEN,
    HTTP_404_NOT_FOUND,
)

from fides.api.api import deps
from fides.api.api.deps import get_config_proxy, get_db
from fides.api.api.v1.endpoints.user_permission_endpoints import validate_user_id
from fides.api.common_exceptions import AuthenticationError
from fides.api.cryptography.cryptographic_util import b64_str_to_str
from fides.api.cryptography.schemas.jwt import JWE_PAYLOAD_CLIENT_ID
from fides.api.models.client import ClientDetail
from fides.api.models.fides_user import FidesUser
from fides.api.models.fides_user_invite import FidesUserInvite
from fides.api.models.fides_user_permissions import FidesUserPermissions
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.api.oauth.roles import APPROVER, VIEWER
from fides.api.oauth.utils import (
    create_temporary_user_for_login_flow,
    extract_payload,
    get_current_user,
    oauth2_scheme,
    verify_oauth_client,
)
from fides.api.schemas.oauth import AccessToken
from fides.api.schemas.user import (
    UserCreate,
    UserCreateResponse,
    UserForcePasswordReset,
    UserLogin,
    UserLoginResponse,
    UserPasswordReset,
    UserResponse,
    UserUpdate,
)
from fides.api.service.deps import get_user_service
from fides.api.util.api_router import APIRouter
from fides.common.api.scope_registry import (
    SCOPE_REGISTRY,
    SYSTEM_MANAGER_DELETE,
    SYSTEM_MANAGER_READ,
    SYSTEM_MANAGER_UPDATE,
    USER_CREATE,
    USER_DELETE,
    USER_PASSWORD_RESET,
    USER_READ,
    USER_UPDATE,
)
from fides.common.api.v1 import urn_registry as urls
from fides.common.api.v1.urn_registry import V1_URL_PREFIX
from fides.config import CONFIG, FidesConfig, get_config
from fides.config.config_proxy import ConfigProxy
from fides.service.user.user_service import UserService

router = APIRouter(tags=["Users"], prefix=V1_URL_PREFIX)


ARTIFICIAL_TEMP_USER = create_temporary_user_for_login_flow(
    CONFIG
)  # To reduce likelihood of timing attacks.  Creating once and holding in memory


def get_system_by_fides_key(db: Session, system_key: FidesKey) -> System:
    """Load a system by FidesKey or throw a 404"""
    system = System.get_by(db, field="fides_key", value=system_key)

    if not system:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"No system found with fides_key {system_key}.",
        )
    return system


def _validate_current_user(user_id: str, user_from_token: FidesUser) -> None:
    if not user_from_token:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"User with ID {user_id} does not exist.",
        )

    if user_id != user_from_token.id:
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED,
            detail="You are only authorised to update your own user data.",
        )


@router.put(
    urls.USER_DETAIL,
    dependencies=[Security(verify_oauth_client)],
    status_code=HTTP_200_OK,
    response_model=UserResponse,
)
async def update_user(
    *,
    db: Session = Depends(deps.get_db),
    authorization: str = Security(oauth2_scheme),
    current_user: FidesUser = Depends(get_current_user),
    user_id: str,
    data: UserUpdate,
) -> FidesUser:
    """
    Update a user given a `user_id`. If the user is not updating their own data,
    they need the USER_UPDATE scope
    """
    user = FidesUser.get(db=db, object_id=user_id)
    if not user:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f"User with id {user_id} not found."
        )

    is_this_user = user.id == current_user.id
    if not is_this_user:
        await verify_oauth_client(
            security_scopes=Security(verify_oauth_client, scopes=[USER_UPDATE]),
            authorization=authorization,
            db=db,
        )

    user.update(db=db, data=data.model_dump(mode="json"))
    logger.info("Updated user with id: '{}'.", user.id)
    return user


@router.post(
    urls.USER_PASSWORD_RESET,
    dependencies=[Security(verify_oauth_client)],
    status_code=HTTP_200_OK,
    response_model=UserResponse,
)
def update_user_password(
    *,
    db: Session = Depends(deps.get_db),
    current_user: FidesUser = Depends(get_current_user),
    user_id: str,
    data: UserPasswordReset,
) -> FidesUser:
    """
    Update a user's password given a `user_id`. By default this is limited to users
    updating their own data.
    """
    _validate_current_user(user_id, current_user)

    if not current_user.credentials_valid(
        b64_str_to_str(data.old_password), CONFIG.security.encoding
    ):
        raise HTTPException(
            status_code=HTTP_401_UNAUTHORIZED, detail="Incorrect password."
        )

    current_user.update_password(db=db, new_password=data.new_password)

    logger.info("Updated user with id: '{}'.", current_user.id)
    return current_user


@router.post(
    urls.USER_FORCE_PASSWORD_RESET,
    dependencies=[Security(verify_oauth_client, scopes=[USER_PASSWORD_RESET])],
    status_code=HTTP_200_OK,
    response_model=UserResponse,
)
def force_update_password(
    *,
    db: Session = Depends(deps.get_db),
    user_id: str,
    data: UserForcePasswordReset,
) -> FidesUser:
    """
    Update any user's password given a `user_id` without needing to know the user's
    previous password.
    """
    user: Optional[FidesUser] = FidesUser.get(db=db, object_id=user_id)
    if not user:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"User with ID {user_id} does not exist.",
        )

    user.update_password(db=db, new_password=data.new_password)
    logger.info("Updated user with id: '{}'.", user.id)
    return user


def logout_oauth_client(
    authorization: str = Security(oauth2_scheme), db: Session = Depends(get_db)
) -> Optional[ClientDetail]:
    """
    Streamlined oauth checks for logout.  Only raises an error if no authorization is supplied.
    Otherwise, regardless if the token is malformed or expired, still return a 204.
    Returns a client if we can extract one from the token.
    """
    if authorization is None:
        raise AuthenticationError(detail="Authentication Failure")

    try:
        token_data = json.loads(
            extract_payload(authorization, CONFIG.security.app_encryption_key)
        )
    except jose.exceptions.JWEParseError:
        return None

    client_id = token_data.get(JWE_PAYLOAD_CLIENT_ID)
    if (
        not client_id or client_id == CONFIG.security.oauth_root_client_id
    ):  # The root client is not a persisted object
        return None

    client = ClientDetail.get(
        db, object_id=client_id, config=CONFIG, scopes=SCOPE_REGISTRY
    )

    return client


@router.post(
    urls.LOGOUT,
    status_code=HTTP_204_NO_CONTENT,
)
def user_logout(
    *,
    client: Optional[ClientDetail] = Security(
        logout_oauth_client,
    ),
    db: Session = Depends(deps.get_db),
) -> None:
    """Logout the user by deleting its client where applicable"""

    logger.info("Logging out user.")
    if client:
        client.delete(db)


@router.put(
    urls.SYSTEM_MANAGER,
    dependencies=[Security(verify_oauth_client, scopes=[SYSTEM_MANAGER_UPDATE])],
    response_model=List[SystemSchema],
)
def update_managed_systems(
    *,
    db: Session = Depends(deps.get_db),
    user_id: str,
    systems: List[FidesKey],
) -> List[SystemSchema]:
    """
    Endpoint to override the systems for which a user is "system manager".
    All systems the user manages are replaced with those in the request body.
    """
    user = validate_user_id(db, user_id)

    if not (user.permissions and user.permissions.roles):  # type: ignore
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"User {user_id} needs permissions before they can be assigned as system manager.",
        )

    if APPROVER in user.permissions.roles:  # type: ignore
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"User {user_id} is an {APPROVER} and cannot be assigned as a system manager.",
        )

    if len(set(systems)) != len(systems):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail=f"Cannot add user {user_id} as system manager. Duplicate systems in request body.",
        )

    retrieved_systems: Query = db.query(System).filter(System.fides_key.in_(systems))
    if retrieved_systems.count() != len(systems):
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Cannot add user {user_id} as system manager. System(s) not found.",
        )

    logger.info("Updating systems for which user {} is system manager", user_id)

    # Adding new systems for which the user is not already a manager
    for system in retrieved_systems:
        if user not in system.data_stewards:
            user.set_as_system_manager(db, system)

    # Removing systems for which the user in no longer a manager
    for system in user.systems.copy():
        if system not in retrieved_systems:
            user.remove_as_system_manager(db, system)

    return user.systems


@router.get(
    urls.SYSTEM_MANAGER,
    response_model=List[SystemSchema],
)
async def get_managed_systems(
    *,
    db: Session = Depends(deps.get_db),
    authorization: str = Security(oauth2_scheme),
    current_user: FidesUser = Depends(get_current_user),
    user_id: str,
) -> List[SystemSchema]:
    """
    Endpoint to retrieve all the systems for which a user is "system manager".
    """
    # A user is able to retrieve their own systems
    if current_user and current_user.id == user_id:
        logger.info(
            "Retrieving current user's {} systems for which they are system manager",
            user_id,
        )
        return current_user.systems

    # User must have a specific scope to be able to read another user's systems
    user = validate_user_id(db, user_id)
    await verify_oauth_client(
        security_scopes=Security(verify_oauth_client, scopes=[SYSTEM_MANAGER_READ]),
        authorization=authorization,
        db=db,
    )
    logger.info("Getting systems for which user {} is system manager", user_id)
    return user.systems


@router.get(
    urls.SYSTEM_MANAGER_DETAIL,
    response_model=SystemSchema,
)
async def get_managed_system_details(
    *,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(deps.get_db),
    user_id: str,
    system_key: FidesKey,
    current_user: FidesUser = Depends(get_current_user),
) -> SystemSchema:
    """
    Endpoint to retrieve a single system managed by the given user.
    """
    system: System = get_system_by_fides_key(db, system_key)

    if current_user and current_user.id == user_id:
        user = current_user
    else:
        await verify_oauth_client(
            security_scopes=Security(verify_oauth_client, scopes=[SYSTEM_MANAGER_READ]),
            authorization=authorization,
            db=db,
        )
        user = validate_user_id(db, user_id)

    if not system in user.systems:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"User {user_id} is not a manager of system {system.fides_key}",
        )

    logger.info(
        "Getting system {} for which user {} is system manager",
        system.fides_key,
        user_id,
    )

    return system


@router.delete(
    urls.SYSTEM_MANAGER_DETAIL,
    dependencies=[Security(verify_oauth_client, scopes=[SYSTEM_MANAGER_DELETE])],
    status_code=HTTP_204_NO_CONTENT,
)
def remove_user_as_system_manager(
    *, db: Session = Depends(deps.get_db), user_id: str, system_key: FidesKey
) -> None:
    """
    Endpoint to remove user as system manager from the given system
    """
    user = validate_user_id(db, user_id)
    system: System = get_system_by_fides_key(db, system_key)

    if not system in user.systems:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"Cannot delete user as system manager. User {user_id} is not a manager of system {system.fides_key}.",
        )

    user.remove_as_system_manager(db, system)
    logger.info("Removed user {} as system manager of {}", user_id, system.fides_key)


@router.post(
    urls.USERS,
    dependencies=[Security(verify_oauth_client, scopes=[USER_CREATE])],
    status_code=HTTP_201_CREATED,
    response_model=UserCreateResponse,
)
def create_user(
    *,
    db: Session = Depends(get_db),
    user_data: UserCreate,
    config_proxy: ConfigProxy = Depends(get_config_proxy),
    user_service: UserService = Depends(get_user_service),
) -> FidesUser:
    """
    Create a user given a username and password.
    If `password` is sent as a base64 encoded string, it will automatically be decoded
    server-side before being encrypted and persisted.
    If `password` is sent as a plaintext string, it will be encrypted and persisted as is.

    The user is given no roles by default.
    """

    # The root user is not stored in the database so make sure here that the user name
    # is not the same as the root user name.
    if (
        config_proxy.security.root_username
        and config_proxy.security.root_username == user_data.username
    ):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Username already exists."
        )

    user = FidesUser.get_by(db, field="username", value=user_data.username)

    if user:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST, detail="Username already exists."
        )

    user = FidesUser.get_by(db, field="email_address", value=user_data.email_address)

    if user:
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="User with this email address already exists.",
        )

    user = FidesUser.create(db=db, data=user_data.model_dump(mode="json"))

    # invite user via email
    user_service.invite_user(user)

    logger.info("Created user with id: '{}'.", user.id)
    FidesUserPermissions.create(
        db=db,
        data={"user_id": user.id, "roles": [VIEWER]},
    )
    return user


@router.delete(
    urls.USER_DETAIL,
    status_code=HTTP_204_NO_CONTENT,
    dependencies=[Security(verify_oauth_client, scopes=[USER_DELETE])],
)
def delete_user(
    *,
    client: ClientDetail = Security(
        verify_oauth_client,
        scopes=[USER_DELETE],
    ),
    db: Session = Depends(get_db),
    user_id: str,
) -> None:
    """Deletes the User and associated ClientDetail if applicable."""
    user = FidesUser.get_by(db, field="id", value=user_id)

    if not user:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND, detail=f"No user found with id {user_id}."
        )

    logger.info("User with id {} deleted by user with id {}", user_id, client.user_id)

    user.delete(db)


@router.get(
    urls.USER_DETAIL,
    dependencies=[Security(verify_oauth_client, scopes=[USER_READ])],
    response_model=UserResponse,
)
def get_user(*, db: Session = Depends(get_db), user_id: str) -> FidesUser:
    """Returns a User based on an Id"""
    user: Optional[FidesUser] = FidesUser.get_by_key_or_id(db, data={"id": user_id})
    if user is None:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="User not found")

    logger.debug("Returning user with id: '{}'.", user_id)
    return user


@router.get(
    urls.USERS,
    dependencies=[Security(verify_oauth_client, scopes=[USER_READ])],
    response_model=Page[UserResponse],
)
def get_users(
    *,
    db: Session = Depends(get_db),
    params: Params = Depends(),
    username: Optional[str] = None,
) -> AbstractPage[FidesUser]:
    """Returns a paginated list of all users"""
    query = FidesUser.query(db)
    if username:
        query = query.filter(FidesUser.username.ilike(f"%{escape_like(username)}%"))

    logger.debug("Returning a paginated list of users.")

    return paginate(query.order_by(FidesUser.created_at.desc()), params=params)


@router.post(
    urls.LOGIN,
    status_code=HTTP_200_OK,
    response_model=UserLoginResponse,
)
def user_login(
    *,
    db: Session = Depends(get_db),
    config: FidesConfig = Depends(get_config),
    user_data: UserLogin,
    user_service: UserService = Depends(get_user_service),
) -> UserLoginResponse:
    """Login the user by creating a client if it doesn't exist, and have that client
    generate a token."""
    user: FidesUser
    client: ClientDetail
    should_raise_exception: bool = False

    if (
        config.security.root_username
        and config.security.root_password
        and config.security.root_username == user_data.username
        and config.security.root_password == user_data.password
    ):
        client_check = ClientDetail.get(
            db,
            object_id=config.security.oauth_root_client_id,
            config=config,
            scopes=config.security.root_user_scopes,
            roles=config.security.root_user_roles,
        )

        if not client_check:
            raise HTTPException(
                status_code=HTTP_404_NOT_FOUND, detail="No root client found."
            )

        # We have already checked for None but mypy still complains. This prevents mypy
        # from complaining.
        client = client_check
        user = FidesUser(
            id=config.security.oauth_root_client_id,
            username=config.security.root_username,
            created_at=datetime.utcnow(),
        )

        logger.warning(
            "Root Username & Password were used to login! If unexpected, review security settings (FIDES__SECURITY__ROOT_USERNAME and FIDES__SECURITY__ROOT_PASSWORD)"
        )

    else:
        user_check: Optional[FidesUser] = FidesUser.get_by(
            db, field="username", value=user_data.username
        )

        if not user_check:
            # Postpone raising the exception to reduce the time differences between
            # login flows for valid and invalid users. Instead, create a temporary user
            # on which we'll perform parallel operations
            should_raise_exception = True
            user_check = ARTIFICIAL_TEMP_USER

        if not user_check.credentials_valid(user_data.password):
            should_raise_exception = True

        # We have already checked for None but mypy still complains. This prevents mypy
        # from complaining.
        user = user_check

        client = user_service.perform_login(
            config.security.oauth_client_id_length_bytes,
            config.security.oauth_client_secret_length_bytes,
            user,
            skip_save=should_raise_exception,
        )

    logger.info("Creating login access token")
    access_code = client.create_access_code_jwe(config.security.app_encryption_key)

    # Sleep for a random time period
    time.sleep(random.uniform(0.00, 0.50))

    if should_raise_exception:
        # Now raise postponed exception!
        raise HTTPException(
            status_code=HTTP_403_FORBIDDEN, detail="Incorrect username or password."
        )

    return UserLoginResponse(
        user_data=user,
        token_data=AccessToken(access_token=access_code),
    )


def verify_invite_code(
    username: str,
    invite_code: str,
    db: Session = Depends(get_db),
) -> FidesUserInvite:
    """
    Security dependency to verify the invite code.
    Returns the validated FidesUserInvite if all the checks pass.
    """
    user_invite = FidesUserInvite.get_by(db, field="username", value=username)

    if not user_invite:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail="User not found.",
        )

    if not user_invite.invite_code_valid(invite_code):
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Invite code is invalid.",
        )

    if user_invite.is_expired():
        raise HTTPException(
            status_code=HTTP_400_BAD_REQUEST,
            detail="Invite code has expired.",
        )

    return user_invite


@router.post(
    urls.USER_ACCEPT_INVITE,
)
def accept_user_invite(
    *,
    db: Session = Depends(get_db),
    user_data: UserForcePasswordReset,
    verified_invite: FidesUserInvite = Depends(verify_invite_code),
    user_service: UserService = Depends(get_user_service),
) -> UserLoginResponse:
    """Sets the password and enables the user if a valid username and invite code are provided."""

    user: Optional[FidesUser] = FidesUser.get_by(
        db=db, field="username", value=verified_invite.username
    )
    if not user:
        raise HTTPException(
            status_code=HTTP_404_NOT_FOUND,
            detail=f"User with username {verified_invite.username} does not exist.",
        )

    user, access_code = user_service.accept_invite(user, user_data.new_password)

    return UserLoginResponse(
        user_data=user,
        token_data=AccessToken(access_token=access_code),
    )



"""merge unified with ctl and ops

Revision ID: 794e5f5b76ae
Revises: 25adddf820a3, c2f7a29c4780
Create Date: 2022-09-02 08:41:29.962231

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "794e5f5b76ae"
down_revision = ("25adddf820a3", "c2f7a29c4780")
branch_labels = None
depends_on = None


def upgrade():
    pass


def downgrade():
    pass



"""add generic email connector types

Revision ID: ed46521679fb
Revises: 587c53fe3e99
Create Date: 2023-05-19 23:25:48.711050

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ed46521679fb"
down_revision = "587c53fe3e99"
branch_labels = None
depends_on = None


def upgrade():
    # Add 'generic_consent_email' and 'generic_erasure_email' to ConnectionType enum
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('mongodb', 'mysql', 'https', 'snowflake', "
        "'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'manual_webhook', "
        "'timescale', 'fides', 'sovrn', 'attentive', 'dynamodb', 'postgres',"
        "'generic_consent_email', 'generic_erasure_email')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")


def downgrade():
    # Remove 'generic_consent_email' and 'generic_erasure_email' from ConnectionType enum
    op.execute(
        "delete from connectionconfig where connection_type in ('generic_consent_email', 'generic_erasure_email')"
    )
    op.execute("alter type connectiontype rename to connectiontype_old")
    op.execute(
        "create type connectiontype as enum('mongodb', 'mysql', 'https', 'snowflake', "
        "'redshift', 'mssql', 'mariadb', 'bigquery', 'saas', 'manual', 'manual_webhook', "
        "'timescale', 'fides', 'sovrn', 'attentive', 'dynamodb', 'postgres')"
    )
    op.execute(
        (
            "alter table connectionconfig alter column connection_type type connectiontype using "
            "connection_type::text::connectiontype"
        )
    )
    op.execute("drop type connectiontype_old")



from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from fastapi import Depends, Security
from fastapi.security import SecurityScopes
from fideslang.models import System as SystemSchema
from loguru import logger
from sqlalchemy.orm import Session

from fides.api.api.deps import get_db
from fides.api.common_exceptions import AuthorizationError
from fides.api.cryptography.schemas.jwt import JWE_PAYLOAD_SYSTEMS
from fides.api.models.client import ClientDetail
from fides.api.models.sql_models import System  # type: ignore[attr-defined]
from fides.api.oauth.system_manager import SYSTEM_MANAGER_SCOPES
from fides.api.oauth.utils import (
    copy_func,
    extract_token_and_load_client,
    has_permissions,
    has_scope_subset,
    oauth2_scheme,
)


class SystemAuthContainer:
    """Data class to store the original request data containing the information to retrieve the system,
    plus the System(s) themselves.

    We're returning the original request body or path param so we can continue to use all the current
    checks that exist for the crud endpoints without modification.  For example, we need to run additional
    checks that the system exists downstream for other purposes.
    """

    def __init__(
        self,
        original_data: Union[str, SystemSchema],
        system: Optional[System],
    ):
        self.original_data = original_data
        self.system = system


def _get_system_from_request_body(
    system_data: SystemSchema, db: Session = Depends(get_db)
) -> SystemAuthContainer:
    """Returns the fetched System (retrieved from the *request body*) along with the original request body.

    This function is passed as a *dependency* into verify_oauth_client_for_system_from_request_body.
    """
    resp = SystemAuthContainer(original_data=system_data, system=None)
    resource_dict = system_data.model_dump(mode="json")
    if resource_dict.get("fides_key"):
        system = (
            db.query(System)
            .filter(System.fides_key == resource_dict.get("fides_key"))
            .first()
        )
        resp.system = system
    return resp


def _get_system_from_fides_key(
    fides_key: str, db: Session = Depends(get_db)
) -> SystemAuthContainer:
    """Returns the fetched System (retrieved from the *fides_key* path parameter) along with the original fides_key.

    This function is passed as a *dependency* into verify_oauth_client_for_system_from_fides_key.
    """
    resp = SystemAuthContainer(original_data=fides_key, system=None)
    system = db.query(System).filter(System.fides_key == fides_key).first()
    resp.system = system
    return resp


async def verify_oauth_client_for_system_from_request_body(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(get_db),
    system_auth_data: SystemAuthContainer = Depends(_get_system_from_request_body),
) -> Union[str, System]:
    """
    Verifies that the access token provided in the authorization header contains the necessary scopes to be
    able to access the System found in the *request body*

    The user must either have 1) the correct scope(s) 2) a role encompassing the necessary scope or 3) the user
    is a system manager of the given system which gives them the proper scope.

    Yields a 403 forbidden error if not.
    """

    system = has_system_permissions(
        system_auth_data=system_auth_data,
        authorization=authorization,
        security_scopes=security_scopes,
        db=db,
    )
    return system


async def verify_oauth_client_for_system_from_fides_key(
    security_scopes: SecurityScopes,
    authorization: str = Security(oauth2_scheme),
    db: Session = Depends(get_db),
    system_auth_data: SystemAuthContainer = Depends(_get_system_from_fides_key),
) -> str:
    """
    Verifies that the access token provided in the authorization header contains the necessary scopes to be
    able to access the System from the given *fides_key* in the path parameter.

    The user must either have 1) the correct scope(s) 2) a role encompassing the necessary scope or 3) the user
    is a system manager of the given system which gives them the proper scope.

    Yields a 403 forbidden error if not.
    """
    system = has_system_permissions(
        system_auth_data=system_auth_data,
        authorization=authorization,
        security_scopes=security_scopes,
        db=db,
    )
    assert isinstance(system, str)
    return system


def has_system_permissions(
    system_auth_data: SystemAuthContainer,
    authorization: str,
    security_scopes: SecurityScopes,
    db: Session,
) -> Union[str, System]:
    """
    Helper method that verifies that the token has the proper permissions to access the system(s).

    This method is shared between methods that use different dependency injections, depending on whether the system
    data is located in the path parameter or in the request body.
    """
    token_data, client = extract_token_and_load_client(
        authorization, db
    )  # Can raise permission errors if issues with client or token

    has_model_level_permissions: bool = (
        has_permissions(  # Token has the correct scope(s) or role(s) associated with the correct scopes
            token_data, client, security_scopes
        )
    )
    has_system_manager_permissions: bool = (
        _has_scope_as_system_manager(  # Token indicates system manager permissions of the current system(s)
            token_data, client, security_scopes, system_auth_data.system
        )
    )

    has_perms: bool = has_model_level_permissions or has_system_manager_permissions

    if not has_perms:
        raise AuthorizationError(detail="Not Authorized for this action")

    return system_auth_data.original_data


def _has_scope_as_system_manager(
    token_data: Dict[str, Any],
    client: ClientDetail,
    endpoint_scopes: SecurityScopes,
    system: Optional[System],
) -> bool:
    """Does the user have the proper scope as system manager?

    Returns True if the 1) system(s) exist, 2) the system(s) are on the user's token, 3) the token is still valid, meaning
    the system(s) are still on the user's client and 4) system manager scopes contain the necessary scope(s) for this endpoint.
    """
    if not system:
        logger.debug("System resource not found.")
        return False

    assigned_systems: List[str] = token_data.get(JWE_PAYLOAD_SYSTEMS, [])
    if not system.id in assigned_systems:
        logger.debug("Systems are not on the user's token.")
        return False

    if not set(assigned_systems).issubset(set(client.systems or [])):
        # If the system on the token is not included on the client, this token is no longer valid
        logger.debug("Client no longer allowed to issue this system.")
        return False

    if not has_scope_subset(
        user_scopes=SYSTEM_MANAGER_SCOPES, endpoint_scopes=endpoint_scopes
    ):
        logger.debug(
            "System manager scopes don't include required scopes to access this endpoint."
        )
        return False

    return True


# This is a workaround so that we can override CLI-related endpoints and
# all other endpoints separately.
verify_oauth_client_for_system_from_request_body_cli = copy_func(
    verify_oauth_client_for_system_from_request_body
)
verify_oauth_client_for_system_from_fides_key_cli = copy_func(
    verify_oauth_client_for_system_from_fides_key
)


async def get_system_schema(system_data: SystemSchema) -> SystemSchema:
    """
    Return system request body directly.  An override to bypass authentication for updating systems in dev mode.
    """
    return system_data


async def get_system_fides_key(fides_key: str) -> str:
    """
    Return system fides_key path parameter directly.  An override to bypass authentication for deleting systems in dev mode.
    """
    return fides_key



"""Add tcf_configuration FK to experience config

Revision ID: 9288f729cac4
Revises: 99c603c1b8f9
Create Date: 2025-04-07 18:49:31.843362

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "9288f729cac4"
down_revision = "99c603c1b8f9"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "privacyexperienceconfig",
        sa.Column("tcf_configuration_id", sa.String(), nullable=True),
    )
    op.create_foreign_key(
        "privacyexperienceconfig_tcf_configuration_fkey",
        "privacyexperienceconfig",
        "tcf_configuration",
        ["tcf_configuration_id"],
        ["id"],
        ondelete="SET NULL",
    )
    op.add_column(
        "privacyexperienceconfighistory",
        sa.Column("tcf_configuration_id", sa.String(), nullable=True),
    )
    op.create_foreign_key(
        "privacyexperienceconfighistory_tcf_configuration_fkey",
        "privacyexperienceconfighistory",
        "tcf_configuration",
        ["tcf_configuration_id"],
        ["id"],
        ondelete="SET NULL",
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(
        "privacyexperienceconfighistory_tcf_configuration_fkey",
        "privacyexperienceconfighistory",
        type_="foreignkey",
    )
    op.drop_column("privacyexperienceconfighistory", "tcf_configuration_id")
    op.drop_constraint(
        "privacyexperienceconfig_tcf_configuration_fkey",
        "privacyexperienceconfig",
        type_="foreignkey",
    )
    op.drop_column("privacyexperienceconfig", "tcf_configuration_id")
    # ### end Alembic commands ###



"""Added messaging template table

Revision ID: 507563f6f8d4
Revises: fd52d5f08c17
Create Date: 2023-08-05 05:30:52.105840

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "507563f6f8d4"
down_revision = "fd52d5f08c17"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "messaging_template",
        sa.Column("id", sa.String(length=255), nullable=False),
        sa.Column(
            "created_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column(
            "updated_at",
            sa.DateTime(timezone=True),
            server_default=sa.text("now()"),
            nullable=True,
        ),
        sa.Column("key", sa.String(), nullable=False),
        sa.Column("content", postgresql.JSONB(astext_type=sa.Text()), nullable=False),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_messagingtemplate_id"), "messaging_template", ["id"], unique=False
    )
    op.create_index(
        op.f("ix_messagingtemplate_key"), "messaging_template", ["key"], unique=True
    )
    # ### end Alembic commands ###


def downgrade():
    op.drop_index(op.f("ix_messagingtemplate_key"), table_name="messaging_template")
    op.drop_index(op.f("ix_messagingtemplate_id"), table_name="messaging_template")
    op.drop_table("messaging_template")



"""Contains the annotate group of CLI commands for fides."""

import rich_click as click

from fides.cli.utils import with_analytics
from fides.core import annotate_dataset as _annotate_dataset


@click.group(name="annotate")
@click.pass_context
def annotate(ctx: click.Context) -> None:
    """
    Interactively annotate Fides resources.
    """


@annotate.command(name="dataset")  # type: ignore
@click.pass_context
@click.argument("input_filename", type=str)
@click.option(
    "-a",
    "--all-members",
    is_flag=True,
    help="Annotate all parts of the dataset including schemas and tables.",
)
@click.option(
    "-v",
    "--validate",
    is_flag=True,
    default=False,
    help="Validate annotation inputs.",
)
@with_analytics
def annotate_dataset(
    ctx: click.Context, input_filename: str, all_members: bool, validate: bool
) -> None:
    """
    Interactively annotate a dataset file in-place.
    """
    config = ctx.obj["CONFIG"]
    _annotate_dataset.annotate_dataset(
        config=config,
        dataset_file=input_filename,
        annotate_all=all_members,
        validate=validate,
    )



"""add property_id indices

Revision ID: 4fba24045fec
Revises: 57479150efd6
Create Date: 2024-05-17 19:25:36.539636

"""

import sqlalchemy as sa
from alembic import op
from loguru import logger

# revision identifiers, used by Alembic.
revision = "4fba24045fec"
down_revision = "57479150efd6"
branch_labels = None
depends_on = None


def upgrade():
    connection = op.get_bind()

    # only run the index creation if the tables have less than 1 million rows
    privacypreferencehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM privacypreferencehistory")
    ).scalar()
    if privacypreferencehistory_count < 1000000:
        op.create_index(
            op.f("ix_privacypreferencehistory_property_id"),
            "privacypreferencehistory",
            ["property_id"],
            unique=False,
        )
    else:
        logger.warning(
            "The privacypreferencehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY ix_privacypreferencehistory_property_id "
            "ON privacypreferencehistory (property_id)'"
        )

    servednoticehistory_count = connection.execute(
        sa.text("SELECT COUNT(*) FROM servednoticehistory")
    ).scalar()
    if servednoticehistory_count < 1000000:
        op.create_index(
            op.f("ix_servednoticehistory_property_id"),
            "servednoticehistory",
            ["property_id"],
            unique=False,
        )
    else:
        logger.warning(
            "The servednoticehistory table has more than 1 million rows, "
            "skipping index creation. Be sure to manually run "
            "'CREATE INDEX CONCURRENTLY ix_servednoticehistory_property_id "
            "ON servednoticehistory (property_id)'"
        )


def downgrade():
    op.drop_index(
        op.f("ix_servednoticehistory_property_id"), table_name="servednoticehistory"
    )
    op.drop_index(
        op.f("ix_privacypreferencehistory_property_id"),
        table_name="privacypreferencehistory",
    )



from __future__ import annotations

import hashlib
from typing import Dict, List, Optional, Type

from fides.api.schemas.masking.masking_configuration import HashMaskingConfiguration
from fides.api.schemas.masking.masking_secrets import (
    MaskingSecretCache,
    MaskingSecretMeta,
    SecretType,
)
from fides.api.schemas.masking.masking_strategy_description import (
    MaskingStrategyConfigurationDescription,
    MaskingStrategyDescription,
)
from fides.api.service.masking.strategy.format_preservation import FormatPreservation
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy
from fides.api.util.encryption.secrets_util import SecretsUtil
from fides.config import CONFIG


class HashMaskingStrategy(MaskingStrategy):
    """Masks a value by hashing it"""

    name = "hash"
    configuration_model = HashMaskingConfiguration

    def __init__(
        self,
        configuration: HashMaskingConfiguration,
    ):
        self.algorithm = configuration.algorithm
        if self.algorithm == HashMaskingConfiguration.Algorithm.SHA_256:
            self.algorithm_function = self._hash_sha256
        elif self.algorithm == HashMaskingConfiguration.Algorithm.SHA_512:
            self.algorithm_function = self._hash_sha512
        self.format_preservation = configuration.format_preservation

    def mask(
        self, values: Optional[List[str]], request_id: Optional[str]
    ) -> Optional[List[Optional[str]]]:
        """Returns the hashed version of the provided values. Returns None if the provided value
        is None"""
        if values is None:
            return None

        masking_meta: Dict[SecretType, MaskingSecretMeta] = (
            self._build_masking_secret_meta()
        )
        salt: str | None = SecretsUtil.get_or_generate_secret(
            request_id,
            SecretType.salt,
            masking_meta[SecretType.salt],
        )

        masked_values: List[Optional[str]] = []
        for value in values:
            if value is None:
                masked_values.append(None)
                continue

            masked: str = self.algorithm_function(str(value), salt)  # type: ignore
            if self.format_preservation is not None:
                formatter = FormatPreservation(self.format_preservation)
                masked = formatter.format(masked)
            masked_values.append(masked)
        return masked_values

    def secrets_required(self) -> bool:
        return True

    def generate_secrets_for_cache(self) -> List[MaskingSecretCache]:
        masking_meta: Dict[SecretType, MaskingSecretMeta] = (
            self._build_masking_secret_meta()
        )
        return SecretsUtil.build_masking_secrets_for_cache(masking_meta)

    # MR Note - We will need a way to ensure that this does not fall out of date. Given that it
    # includes subjective instructions, this is not straightforward to automate
    @classmethod
    def get_description(cls: Type[MaskingStrategy]) -> MaskingStrategyDescription:
        return MaskingStrategyDescription(
            name=cls.name,
            description="Masks the input value by returning a hashed version of the input value",
            configurations=[
                MaskingStrategyConfigurationDescription(
                    key="algorithm",
                    description="Specifies the hashing algorithm to be used. Can be SHA-256 or "
                    "SHA-512. If not provided, default is SHA-256",
                ),
                MaskingStrategyConfigurationDescription(
                    key="format_preservation",
                    description="Option to preserve format in masking, with a provided suffix",
                ),
            ],
        )

    @staticmethod
    def data_type_supported(data_type: Optional[str]) -> bool:
        """Determines whether or not the given data type is supported by this masking strategy"""
        supported_data_types = {"string"}
        return data_type in supported_data_types

    # Helpers

    @staticmethod
    def _hash_sha256(value: str, salt: str) -> str:
        return hashlib.sha256(
            (value + salt).encode(CONFIG.security.encoding)
        ).hexdigest()

    @staticmethod
    def _hash_sha512(value: str, salt: str) -> str:
        return hashlib.sha512(
            (value + salt).encode(CONFIG.security.encoding)
        ).hexdigest()

    @classmethod
    def _build_masking_secret_meta(
        cls: Type[MaskingStrategy],
    ) -> Dict[SecretType, MaskingSecretMeta]:
        return {
            SecretType.salt: MaskingSecretMeta[str](
                masking_strategy=cls.name,
                generate_secret_func=SecretsUtil.generate_secret_string,
            )
        }



import string
from secrets import choice
from typing import List, Optional, Type

from fides.api.schemas.masking.masking_configuration import (
    RandomStringMaskingConfiguration,
)
from fides.api.schemas.masking.masking_strategy_description import (
    MaskingStrategyConfigurationDescription,
    MaskingStrategyDescription,
)
from fides.api.service.masking.strategy.format_preservation import FormatPreservation
from fides.api.service.masking.strategy.masking_strategy import MaskingStrategy


class RandomStringRewriteMaskingStrategy(MaskingStrategy):
    """Masks each provied value with a random string of the length specified in the configuration."""

    name = "random_string_rewrite"
    configuration_model = RandomStringMaskingConfiguration

    def __init__(
        self,
        configuration: RandomStringMaskingConfiguration,
    ):
        self.length = configuration.length
        self.format_preservation = configuration.format_preservation

    def mask(
        self, values: Optional[List[str]], request_id: Optional[str]
    ) -> Optional[List[str]]:
        """Replaces the value with a random lowercase string of the configured length"""
        if values is None:
            return None

        masked_values: List[str] = []
        for _ in range(len(values)):
            masked: str = "".join(
                [
                    choice(string.ascii_lowercase + string.digits)
                    for _ in range(self.length)
                ]
            )
            if self.format_preservation is not None:
                formatter = FormatPreservation(self.format_preservation)
                masked = formatter.format(masked)
            masked_values.append(masked)
        return masked_values

    def secrets_required(self) -> bool:
        return False

    @classmethod
    def get_description(cls: Type[MaskingStrategy]) -> MaskingStrategyDescription:
        return MaskingStrategyDescription(
            name=cls.name,
            description="Masks the input value with a random string of a specified length",
            configurations=[
                MaskingStrategyConfigurationDescription(
                    key="length",
                    description="Specifies the desired length of the randomized string.",
                )
            ],
        )

    @staticmethod
    def data_type_supported(data_type: Optional[str]) -> bool:
        """Determines whether or not the given data type is supported by this masking strategy"""
        supported_data_types = {"string"}
        return data_type in supported_data_types



from typing import Optional

from fideslang.validation import FidesKey
from pydantic import ConfigDict

from fides.api.schemas.base_class import FidesSchema
from fides.api.schemas.connection_configuration.connection_config import (
    ConnectionConfigurationResponse,
)


class WebhookBase(FidesSchema):
    """Base schema for Webhooks"""

    key: Optional[FidesKey] = None
    name: str


class PreApprovalWebhookCreate(WebhookBase):
    """Request schema for creating/updating a Pre Approval Webhook"""

    connection_config_key: FidesKey
    model_config = ConfigDict(use_enum_values=True)


class PreApprovalWebhookResponse(WebhookBase):
    """Response schema after creating/updating/getting a PreApprovalWebhook"""

    connection_config: Optional[ConnectionConfigurationResponse] = None
    model_config = ConfigDict(from_attributes=True)


class PreApprovalWebhookUpdate(FidesSchema):
    """Request schema for updating a single webhook - fields are optional"""

    name: Optional[str] = None
    connection_config_key: Optional[FidesKey] = None
    model_config = ConfigDict(
        from_attributes=True, extra="forbid", use_enum_values=True
    )



"""Rename user name tables and indexes

Revision ID: c7cc36820d4b
Revises: ed1b00ff963d
Create Date: 2022-06-25 15:20:20.373137

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "c7cc36820d4b"
down_revision = "ed1b00ff963d"
branch_labels = None
depends_on = None


def upgrade():
    op.rename_table("fidesopsuser", "fidesuser")
    op.execute("ALTER INDEX fidesopsuser_pkey RENAME TO fidesuser_pkey")
    op.execute("ALTER INDEX ix_fidesopsuser_id RENAME TO ix_fidesuser_id")
    op.execute("ALTER INDEX ix_fidesopsuser_username RENAME TO ix_fidesuser_username")
    op.rename_table("fidesopsuserpermissions", "fidesuserpermissions")
    op.execute(
        "ALTER INDEX fidesopsuserpermissions_pkey RENAME TO fidesuserpermission_pkey"
    )
    op.execute(
        "ALTER INDEX fidesopsuserpermissions_user_id_key RENAME TO fidesuserpermissions_user_id_key"
    )
    op.execute(
        "ALTER INDEX ix_fidesopsuserpermissions_id RENAME TO ix_fidesuserpermissions_id"
    )
    op.drop_constraint(
        constraint_name="fidesopsuserpermissions_user_id_fkey",
        table_name="fidesuserpermissions",
        type_="foreignkey",
    )
    op.create_foreign_key(
        constraint_name="fidesuserpermissions_user_id_fkey",
        source_table="fidesuserpermissions",
        referent_table="fidesuser",
        local_cols=["user_id"],
        remote_cols=["id"],
    )
    op.drop_constraint(
        constraint_name="client_user_id_fkey", table_name="client", type_="foreignkey"
    )
    op.create_foreign_key(
        constraint_name="client_user_id_fkey",
        source_table="client",
        referent_table="fidesuser",
        local_cols=["user_id"],
        remote_cols=["id"],
    )
    op.drop_constraint(
        constraint_name="privacyrequest_reviewed_by_fkey",
        table_name="privacyrequest",
        type_="foreignkey",
    )
    op.create_foreign_key(
        constraint_name="privacyrequest_reviewed_by_fkey",
        source_table="privacyrequest",
        referent_table="fidesuser",
        local_cols=["reviewed_by"],
        remote_cols=["id"],
    )
    op.drop_constraint(
        "privacyrequest_reviewed_by_fkey", "privacyrequest", type_="foreignkey"
    )
    op.create_foreign_key(
        None,
        "privacyrequest",
        "fidesuser",
        ["reviewed_by"],
        ["id"],
        ondelete="SET NULL",
    )


def downgrade():
    op.rename_table("fidesuser", "fidesopsuser")
    op.execute("ALTER INDEX fidesuer_pkey RENAME TO fidesopsuser_pkey")
    op.execute("ALTER INDEX ix_fidesuer_id RENAME TO ix_fidesopsuser_id")
    op.execute("ALTER INDEX ix_fidesuer_username RENAME TO ix_fidesopsuser_username")
    op.rename_table("fidesuserpermissions", "fidesopsuserpremissions")
    op.execute(
        "ALTER INDEX fidesuerpermission_pkey RENAME TO fidesopsuserpermissions_pkey"
    )
    op.execute(
        "ALTER INDEX fidesuerpermissions_user_id_key RENAME TO fidesopsuserpermissions_user_id_key"
    )
    op.execute(
        "ALTER INDEX ix_fidesuserpermissions_id RENAME TO ix_fidesopsuserpermissions_id"
    )
    op.drop_constraint(
        constraint_name="fidesuserpermissions_user_id_fkey",
        table_name="fidesopsuserpermissions",
        type_="foreignkey",
    )
    op.create_foreign_key(
        constraint_name="fidesopsuserpermissions_user_id_fkey",
        source_table="fidesopsuserpermissions",
        referent_table="fidesopsuser",
        local_cols=["user_id"],
        remote_cols=["id"],
    )
    op.drop_constraint(
        constraint_name="client_user_id_fkey", table_name="client", type_="foreignkey"
    )
    op.create_foreign_key(
        constraint_name="client_user_id_fkey",
        source_table="client",
        referent_table="fidesopsuser",
        local_cols=["user_id"],
        remote_cols=["id"],
    )
    op.drop_constraint(
        constraint_name="privacyrequest_reviewed_by_fkey",
        table_name="privacyrequest",
        type_="foreignkey",
    )
    op.create_foreign_key(
        constraint_name="privacyrequest_reviewed_by_fkey",
        source_table="privacyrequest",
        referent_table="fidesopsuser",
        local_cols=["reviewed_by"],
        remote_cols=["id"],
    )
    op.drop_constraint(None, "privacyrequest", type_="foreignkey")
    op.create_foreign_key(
        "privacyrequest_reviewed_by_fkey",
        "privacyrequest",
        "fidesuser",
        ["reviewed_by"],
        ["id"],
    )
